from HPO.data.HAR import *
from HPO.data.SHAR import *
from HPO.data.TEPS import *
from HPO.data.EEG import *
from HPO.data.FORDB import *
from HPO.data.UEA_datasets import *
from Transformers.datasets.androzoo_dl import *
from Transformers.datasets import *
DATASETS = {
    "teps" : (Train_TEPS,Test_TEPS),
    "HAR" : (Train_HAR,Test_HAR),
    #"Full_SHAR" : (Full_SHAR, Full_SHAR),
    "EEG" : (Train_EEG,Test_EEG),
    "EEG_Retrain" : (Train_EEG,Test_EEG),
    "LSST" : (Train_LSST,Test_LSST),
    "PhonemeSpectra" : (Train_PhonemeSpectra,Test_PhonemeSpectra),
    "FaceDetection" : (Train_FaceDetection,Test_FaceDetection),
    "PenDigits" : (Train_PenDigits,Test_PenDigits),
    "PenDigitsRetrain" : (Train_PenDigits,Validation_PenDigits),
    "FORDB" : (Train_FORDB,Test_FORDB),
    "FaceDetection" : (Train_FaceDetection,Test_FaceDetection),
    "CharacterTrajectories" : (Train_CharacterTrajectories,Test_CharacterTrajectories),
    "SpokenArabicDigits" : (Train_SpokenArabicDigits,Test_SpokenArabicDigits),
    "FaceDetectionRetrain" : (Train_FaceDetection,Validation_FaceDetection),
    "PhonemeSpectraRetrain" : (Train_PhonemeSpectra,Validation_PhonemeSpectra),
    "EthanolConcentration" : (Train_EthanolConcentration,Test_EthanolConcentration),
    
    "LSST_Retrain" : (Train_LSST,Validation_LSST),
    "Hex" : (Train_Hex,Test_Hex),
    "FaceDetectionVal" : (Train_FaceDetection,Validation_FaceDetection),
    "FaceDetectionTest" : (Train_FaceDetection,True_Test_FaceDetection),
    "PenDigitsVal" : (Train_PenDigits,Validation_PenDigits),
    "PenDigitsTest" : (Train_PenDigits,True_Test_PenDigits),
    "Full_FaceDetection" : (Full_FaceDetection,Full_FaceDetection),
    "Full_SpokenArabicDigits" : (Full_SpokenArabicDigits,Full_SpokenArabicDigits),

    "Full_LSST" : (Full_LSST,Full_LSST),
    "Full_EthanolConcentration" : (Full_EthanolConcentration,Full_EthanolConcentration),
    "Full_PenDigits" : (Full_PenDigits,Full_PenDigits),
    "Full_PhonemeSpectra" : (Full_PhonemeSpectra,Full_PhonemeSpectra),
    "UWaveGestureLibrary" : (Train_UWaveGestureLibrary,Test_UWaveGestureLibrary),
    "Full_UWaveGestureLibrary" : (Full_UWaveGestureLibrary,Full_UWaveGestureLibrary)
    
}


def get_dataset(name,train_args,test_args):
    if name in DATASETS:
        if test_args == None:
            return DATASETS[name][0](**train_args)
        return DATASETS[name][0](**train_args),DATASETS[name][1](**test_args)
    else:

        if "Full" in name:
            return Full_N(name.split("_")[-1],**train_args)
        elif "Retrain" in name:
            return Retrain_N(name.split("_")[0],**train_args), Validation_N(name.split("_")[0],**test_args)
        if test_args == None:
            return Train_N(name,**train_args)
        else:
            return Train_N(name,**train_args), Test_N(name,**test_args)




import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.cifar100 import CIFAR100_Train, CIFAR100_Test
import time
import sys
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model, auto_train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)


def _compute(hyperparameter,cuda_device, JSON_CONFIG ):
  ### Configuration 
  with open(JSON_CONFIG) as f:
    data = json.load(f)
    SETTINGS = data["WORKER_CONFIG"]
    SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  
  if cuda_device == None:
     cuda_device = 3
  torch.cuda.empty_cache()

  torch.cuda.set_device(cuda_device)
  torch.set_float32_matmul_precision('high')
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:
  for i in range(1):
      ##Dataset Initialisation
      #datasets = UEA_Handler("/home/cmackinnon/scripts/datasets/UEA/")
      name = SETTINGS["DATASET_CONFIG"]["NAME"]
      #train_args = [False, cuda_device ,None,1]
      # test_args = [False, cuda_device , None,1]
      if "AUGMENTATIONS" in SETTINGS:
        augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
      else: 
        augs = None 
      train_dataset = CIFAR100_Train(cuda_device)
      test_dataset = CIFAR100_Test(cuda_device)
      #test_dataset = datasets.load_all(name,train_args,test_args)

      
      print("Cuda Device Value: ", cuda_device)

      n_classes = train_dataset.get_n_classes()
      multibatch = False
      torch.cuda.empty_cache()
      trainloader = torch.utils.data.DataLoader(
                              train_dataset,shuffle = True,
                              batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          test_dataset,shuffle = True,
                          batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
      n_classes = test_dataset.get_n_classes()
      evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
      print("classes: {}".format(train_dataset.get_n_classes()))
      g = GraphConfigSpace(150)
      s = g.sample_configuration()
      s = s[0]
      model = ModelGraph(train_dataset.get_n_features(),32,train_dataset.get_n_classes(),217,s["graph"],s["ops"],device = cuda_device,data_dim =2)

      model = model.cuda(device = cuda_device)
      model = torch.compile(model)
      #torch._dynamo.config.verbose=True
      #explanation, out_guards, graphs, ops_per_graph,break_reasons, explanation_verbose = torch._dynamo.explain(model, torch.rand(16,3,32,32).cuda(3))
      """
      ### Train the model
      """
      #print(explanation_verbose)
      params = sum(p.numel() for p in model.parameters() if p.requires_grad)
      print("Size: {}".format(params))
      train_model(model , SETTINGS, trainloader , cuda_device,logger = False, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None) 
    #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
  torch.cuda.empty_cache()
  model.eval()
  evaluator.forward_pass(model, testloader,SETTINGS["BINARY"])
  evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"])
  total = evaluator.T()
  acc  =  evaluator.T_ACC()
  recall = evaluator.TPR(1)
  recall_total = evaluator.P(1)
  print("Accuracy: ", "%.4f" % ((acc)*100), "%")
  print("Recall: ", "%.4f" % ((recall)*100), "%")
  torch.save(model.state_dict(),"{}/weights/{:.02f}-{}".format(SAVE_PATH,acc,hyperparameter["ID"]))
  return acc, recall,params


if __name__ == "__main__":
  for i in range(500):
    with open(sys.argv[1]) as f:
      HP = json.load(f)["WORKER_CONFIG"]
      HP["ID"] = i
    _compute(HP,2,sys.argv[1])



import torch
import torch.nn as nn




# input - > training loss , training accuracy, validation accuracy, validation loss

#output -> augmentation params , lr 

 
class AutoTrainer(nn.Module):
  def __init__(self,channels):
    super(AutoTrainer,self).__init__()

    self.channels = channels
    self.stem = nn.Conv1d(2 , channels[0] , 3)
    self.GAP = nn.AdaptiveAvgPool1d(1)
    self.fc = nn.Linear(self.channels, 1)
    self.act = nn.ReLU()
    self.fcact = nn.Softmax()
    optimizer = torch.optim.Adam(self.parameters(),lr = 0.0025)
 
  def forward(self, x):
    x = self.stem(x)
    x = self.GAP(x)
    x = torch.flatten(x, 1)
    return self.fcact(self.fc(x))


  def update(self,loss):
    



import pickle

def save_obj(obj, name ):
    with open( name + '.pkl', 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

def load_obj(name ):
    with open( name + '.pkl', 'rb') as f:
        return pickle.load(f)



from HPO.utils.seed import set_seed
import torch
from queue import Empty
from HPO.data.dataset import get_dataset
import HPO.utils.augmentation as aug
from sklearn.model_selection import StratifiedKFold as KFold
import json
import time
def __compute( ID, configs , gpus , res   , JSON_CONFIG, _compute):
  set_seed(JSON_CONFIG)
  device = None
  config = None
  print("Starting process: {}".format(ID))
  train, test = None, None
  WAIT = True
  while not configs.empty() or WAIT:
    if configs.empty():
      time.sleep(1)
      WAIT = False
      continue
    else:
      WAIT = True

    try:
      if device == None:
        device = gpus.get(timeout = 10)
      config = configs.get(timeout = 10)
    except Empty:
      if device != None:
        gpus.put(device)
        return
      
    except:
      torch.cuda.empty_cache()
      if device != None:
        gpus.put(device)
      return
    if device != None:
      if (train == None) and (test == None):
        train, test = load_dataset(JSON_CONFIG,device)
      elif device != train.x.get_device():
        train.update_device(device)
        test.update_device(device)
      acc , rec, params =  _compute(hyperparameter = config , cuda_device = device,JSON_CONFIG = JSON_CONFIG,train_dataset = train, test_dataset= test)
      res.put([config , acc , rec ,params]) 

  torch.cuda.empty_cache()
  print("Got out of configs.empty loop")
  return True



def load_dataset(JSON_CONFIG, cuda_device):
  if type(JSON_CONFIG) != dict:
      with open(JSON_CONFIG) as f:
        data = json.load(f)
  else:
      data = JSON_CONFIG
  SETTINGS = data["WORKER_CONFIG"]

  ##Dataset Initialisation
  name = SETTINGS["DATASET_CONFIG"]["NAME"]
  if data["GENERATE_PARTITION"]:
    DS_PATH = SETTINGS["DATASET_CONFIG"]["DATASET_PATH"]
  else:
    DS_PATH = None
  if "AUGMENTATIONS" in SETTINGS:
    augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
  else: 
    augs = None

  train_args = {"cuda_device":cuda_device,"augmentation" : augs, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}
  test_args = {"cuda_device":cuda_device,"augmentation" :None, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}

  if SETTINGS["RESAMPLES"]:
    dataset = get_dataset(name,train_args,None )
    kfold = KFold(n_splits = 2, shuffle = True)
    splits = [(None,None)]*SETTINGS["RESAMPLES"]
    train_dataset = dataset
    test_dataset = dataset
  elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
    train_dataset, test_dataset = get_dataset(name,train_args, test_args)
    splits = [(None,None)]
  elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
    dataset, test_dataset = get_dataset(name,train_args, test_args)
    kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = False)
    splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())
    train_dataset = dataset
    test_dataset = dataset

  return train_dataset, test_dataset



import math
import random
from HPO.utils.worker_helper import collate_fn_padd
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.genotypes import NASNet
from HPO.utils.DARTS_utils import config_space_2_DARTS
import torch
import numpy as np
import seaborn as sns
from HPO.searchspaces import DARTS_config
import csv
from HPO.data.datasets import Train_TEPS_split

device = 0
net_K = 0
scores = []
def load_csv(file):

  acc_l = []
  rec_l = []
  config_l = []
  with open(file, "r") as csvfile:
    reader = csv.reader(csvfile)
    for row in reader:
      acc = float(row[0])
      rec = float(row[1])
      config = eval(row[2])
      acc_l.append(acc)
      rec_l.append(rec)
      config_l.append(config)

  return acc_l, rec_l , config_l

def hooklogdet(K, labels=None):
    s, ld = np.linalg.slogdet(K)
    return ld

def get_batch_jacobian(net, x, target, device, args=None):
    net.zero_grad()
    x.requires_grad_(True)
    y = net(x)
    y.backward(torch.ones_like(y))
    jacob = x.grad.detach()
    return jacob, target.detach(), y.detach()


##Hooks from author implimentation
def counting_forward_hook(module, inp, out):
        global net_K
        if isinstance(inp, tuple):
            inp = inp[0]
        inp = inp.view(inp.size(0), -1)
        x = (inp > 0).float()
        K = x @ x.t()
        K2 = (1.-x) @ (1.-x.t())
        net_K = net_K + K.cpu().numpy() + K2.cpu().numpy()
    
def counting_backward_hook(module, inp, out):
    module.visited_backwards = True

acc, rec, conf = load_csv("/home/snaags/uist/RegEvo.csv")
acc_list = []
dataset = Train_TEPS_split()
N = 256
N_iter = 8
trainloader = torch.utils.data.DataLoader(
                    dataset,collate_fn = collate_fn_padd, 
                    batch_size=N, drop_last = True)
configspace = DARTS_config.init_config()
###Loop through random models in searchspace
for _ in range(100):
    c = configspace.sample_configuration()
    print("Model: {}".format(_))
    gen = config_space_2_DARTS(c)
    model = NetworkMain(52,c["channels"],2,3,False,0,gen) ##Initialise Network
    model.cuda()

    ##Attach Activation Hooks
    x = repr(model)
    for i in model.cells:
        if "FactorizedReduce" not in repr(i.preprocess0) :
          for x in i.preprocess0.op:
            if repr(x) == "ReLU()":
                x.register_backward_hook(counting_backward_hook)
                x.register_forward_hook(counting_forward_hook)
          for x in i.preprocess1.op:
            if repr(x) == "ReLU()":
                x.register_backward_hook(counting_backward_hook)
                x.register_forward_hook(counting_forward_hook)
        #print("Activation Functions: {} {}".format(act1, act2))
    for j in i._ops:
      if "Conv" in repr(j) and "FactorizedReduce" not in repr(j):
        for j_i in j.op:
          if repr(j_i ) == "ReLU()":
            j_i.register_backward_hook(counting_backward_hook)
            j_i.register_forward_hook(counting_forward_hook)

    net_K = torch.zeros(N,N)

    for i in range(N_iter):
      x, target = next(iter(trainloader))
      x ,target = x.to(device).float(), target.to(device)
      model(x)
    scores.append(hooklogdet(net_K))
    print(scores[-1])
print(scores)
print(acc)
plt.plot(scores)
plt.show()      



import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
import os
import csv
import time 
from HPO.utils.ConfigStruct import Parameter, Cumulative_Integer_Struct, LTP_Parameter 

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""
def get_ops():
  return [ 
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'max_pool_31x31',
    'avg_pool_31x31',
    'max_pool_64x64',
    'avg_pool_64x64',
    'skip_connect',
    'point_conv' ,
    'depth_conv_7',
    'depth_conv_15',
    'depth_conv_29' ,
    'depth_conv_61' ,
    'depth_conv_101',
    'depth_conv_201',
    'SE_8',
    'channel_mlp',
    'SE_16',
    'attention_space',
    'attention_channel']

def init_config():

  cs = CS.ConfigurationSpace()

  conv_ops= [ 
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'max_pool_31x31',
    'avg_pool_31x31',
    'max_pool_64x64',
    'avg_pool_64x64',
    'skip_connect',
    'point_conv' ,
    'depth_conv_7',
    'depth_conv_15',
    'depth_conv_29' ,
    'depth_conv_61' ,
    'depth_conv_101',
    'depth_conv_201',
    'SE_8',
    'SE_16',
    'attention_space',
    'attention_channel']
    #'patch_61',
    #'patch_29'] 
    #'patch_61']
  

  ###DARTS###
  normal_node_0_0 = CSH.CategoricalHyperparameter('normal_node_0_0', choices=conv_ops)
  normal_node_0_1 = CSH.CategoricalHyperparameter('normal_node_0_1', choices=conv_ops)
  normal_index_0_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_0", lower = 0, upper = 1)
  normal_index_0_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_1", lower = 0, upper = 1)
  #normal_index_0_2 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_2", lower = 0, upper = 1)

  normal_node_1_0 = CSH.CategoricalHyperparameter('normal_node_1_0', choices=conv_ops)
  normal_node_1_1 = CSH.CategoricalHyperparameter('normal_node_1_1', choices=conv_ops)
  normal_index_1_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_0", lower = 0, upper = 2)
  normal_index_1_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_1", lower = 0, upper = 2)
  #normal_index_1_2 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_2", lower = 0, upper = 2)

  normal_node_2_0 = CSH.CategoricalHyperparameter('normal_node_2_0', choices=conv_ops)
  normal_node_2_1 = CSH.CategoricalHyperparameter('normal_node_2_1', choices=conv_ops)
  normal_index_2_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_0", lower = 0, upper = 3)
  normal_index_2_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_1", lower = 0, upper = 3)
  #normal_index_2_2 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_2", lower = 0, upper = 3)

  normal_node_3_0 = CSH.CategoricalHyperparameter('normal_node_3_0', choices=conv_ops)
  normal_node_3_1 = CSH.CategoricalHyperparameter('normal_node_3_1', choices=conv_ops)
  normal_index_3_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_0", lower = 0, upper = 4)
  normal_index_3_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_1", lower = 0, upper = 4)
  #normal_index_3_2 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_2", lower = 0, upper = 4)

  reduction_node_0_0 = CSH.CategoricalHyperparameter('reduction_node_0_0', choices=conv_ops)
  reduction_node_0_1 = CSH.CategoricalHyperparameter('reduction_node_0_1', choices=conv_ops)
  reduction_index_0_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_0_0", lower = 0, upper = 1)
  reduction_index_0_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_0_1", lower = 0, upper = 1)
  #reduction_index_0_2 = CSH.UniformIntegerHyperparameter(name = "reduction_index_0_2", lower = 0, upper = 1)

  reduction_node_1_0 = CSH.CategoricalHyperparameter('reduction_node_1_0', choices=conv_ops)
  reduction_node_1_1 = CSH.CategoricalHyperparameter('reduction_node_1_1', choices=conv_ops)
  reduction_index_1_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_1_0", lower = 0, upper = 2)
  reduction_index_1_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_1_1", lower = 0, upper = 2)
  #reduction_index_1_2 = CSH.UniformIntegerHyperparameter(name = "reduction_index_1_2", lower = 0, upper = 2)

  reduction_node_2_0 = CSH.CategoricalHyperparameter('reduction_node_2_0', choices=conv_ops)
  reduction_node_2_1 = CSH.CategoricalHyperparameter('reduction_node_2_1', choices=conv_ops)
  reduction_index_2_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_2_0", lower = 0, upper = 3)
  reduction_index_2_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_2_1", lower = 0, upper = 3)
  #reduction_index_2_2 = CSH.UniformIntegerHyperparameter(name = "reduction_index_2_2", lower = 0, upper = 3)

  reduction_node_3_0 = CSH.CategoricalHyperparameter('reduction_node_3_0', choices=conv_ops)
  reduction_node_3_1 = CSH.CategoricalHyperparameter('reduction_node_3_1', choices=conv_ops)
  reduction_index_3_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_3_0", lower = 0, upper = 4)
  reduction_index_3_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_3_1", lower = 0, upper = 4)
  #reduction_index_3_2 = CSH.UniformIntegerHyperparameter(name = "reduction_index_3_2", lower = 0, upper = 4)


#####################################################


  normal_node_4_0 = CSH.CategoricalHyperparameter('normal_node_4_0', choices=conv_ops)
  normal_node_4_1 = CSH.CategoricalHyperparameter('normal_node_4_1', choices=conv_ops)
  normal_index_4_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_4_0", lower = 0, upper = 5)
  normal_index_4_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_4_1", lower = 0, upper = 5)

  normal_node_5_0 = CSH.CategoricalHyperparameter('normal_node_5_0', choices=conv_ops)
  normal_node_5_1 = CSH.CategoricalHyperparameter('normal_node_5_1', choices=conv_ops)
  normal_index_5_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_5_0", lower = 0, upper = 6)
  normal_index_5_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_5_1", lower = 0, upper = 6)

  normal_node_6_0 = CSH.CategoricalHyperparameter('normal_node_6_0', choices=conv_ops)
  normal_node_6_1 = CSH.CategoricalHyperparameter('normal_node_6_1', choices=conv_ops)
  normal_index_6_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_6_0", lower = 0, upper = 7)
  normal_index_6_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_6_1", lower = 0, upper = 7)

  normal_node_7_0 = CSH.CategoricalHyperparameter('normal_node_7_0', choices=conv_ops)
  normal_node_7_1 = CSH.CategoricalHyperparameter('normal_node_7_1', choices=conv_ops)
  normal_index_7_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_7_0", lower = 0, upper = 8)
  normal_index_7_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_7_1", lower = 0, upper = 8)

  reduction_node_4_0 = CSH.CategoricalHyperparameter('reduction_node_4_0', choices=conv_ops)
  reduction_node_4_1 = CSH.CategoricalHyperparameter('reduction_node_4_1', choices=conv_ops)
  reduction_index_4_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_4_0", lower = 0, upper = 5)
  reduction_index_4_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_4_1", lower = 0, upper = 5)

  reduction_node_5_0 = CSH.CategoricalHyperparameter('reduction_node_5_0', choices=conv_ops)
  reduction_node_5_1 = CSH.CategoricalHyperparameter('reduction_node_5_1', choices=conv_ops)
  reduction_index_5_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_5_0", lower = 0, upper = 6)
  reduction_index_5_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_5_1", lower = 0, upper = 6)

  reduction_node_6_0 = CSH.CategoricalHyperparameter('reduction_node_6_0', choices=conv_ops)
  reduction_node_6_1 = CSH.CategoricalHyperparameter('reduction_node_6_1', choices=conv_ops)
  reduction_index_6_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_6_0", lower = 0, upper = 7)
  reduction_index_6_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_6_1", lower = 0, upper = 7)

  reduction_node_7_0 = CSH.CategoricalHyperparameter('reduction_node_7_0', choices=conv_ops)
  reduction_node_7_1 = CSH.CategoricalHyperparameter('reduction_node_7_1', choices=conv_ops)
  reduction_index_7_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_7_0", lower = 0, upper = 8)
  reduction_index_7_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_7_1", lower = 0, upper = 8)

  layers = CSH.UniformIntegerHyperparameter(name = "layers", lower = 3, upper = 9)
  channels = CSH.UniformIntegerHyperparameter(name = "channels", lower = 3, upper = 7)
  cut_mix_rate= CSH.UniformFloatHyperparameter(name = "cut_mix_rate",lower = 0.0  ,upper = 2)


    ###Topology Definition]###



    ###Topology Definition]###
  
  hp_list = [
        layers,
        channels,
        normal_node_0_0 ,
        normal_node_0_1 ,
        normal_index_0_0,

        normal_index_0_1,
        #normal_index_0_2,
        normal_node_1_0 ,

        normal_node_1_1 ,
        normal_index_1_0,
        normal_index_1_1,
        #normal_index_1_2,

        normal_node_2_0 ,
        normal_node_2_1 ,
        normal_index_2_0,
        normal_index_2_1,
        #normal_index_2_2,
        
        normal_node_3_0 ,
        normal_node_3_1 ,
        normal_index_3_0, 
        normal_index_3_1,
        #normal_index_3_2,
        reduction_node_0_0 ,
        reduction_node_0_1 ,
        reduction_index_0_0,
        reduction_index_0_1,
        #reduction_index_0_2,
        reduction_node_1_0 ,
        reduction_node_1_1 ,
        reduction_index_1_0,
        reduction_index_1_1,
        #reduction_index_1_2,
        reduction_node_2_0 ,
        reduction_node_2_1 ,
        reduction_index_2_0,
        reduction_index_2_1,
        #reduction_index_2_2,
        reduction_node_3_0 ,
        reduction_node_3_1 ,
        reduction_index_3_0, 
        reduction_index_3_1,
        #reduction_index_3_2
        normal_node_4_0 ,
        normal_node_4_1 ,
        normal_index_4_0,

        normal_index_4_1,
        normal_node_5_0 ,

        normal_node_5_1 ,
        normal_index_5_0,

        normal_index_5_1,
        normal_node_6_0 ,

        normal_node_6_1 ,
        normal_index_6_0,

        normal_index_6_1,
        normal_node_7_0 ,

        normal_node_7_1 ,
        normal_index_7_0, 
        normal_index_7_1,
        reduction_node_4_0 ,
        reduction_node_4_1 ,
        reduction_index_4_0,

        reduction_index_4_1,
        reduction_node_5_0 ,

        reduction_node_5_1 ,
        reduction_index_5_0,

        reduction_index_5_1,
        reduction_node_6_0 ,

        reduction_node_6_1 ,
        reduction_index_6_0,

        reduction_index_6_1,
        reduction_node_7_0 ,

        reduction_node_7_1 ,
        reduction_index_7_0, 
        reduction_index_7_1,

]

  cs.add_hyperparameters(hp_list)
  return cs

if __name__ == "__main__":
  from HPO.utils.DARTS_utils import config_space_2_DARTS
  configS = init_config()
  print(configS.get_hyperparameters())
  c = configS.sample_configuration()
  print(c)
  print(config_space_2_DARTS(c))



import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
import os
import csv
import time 
from HPO.utils.ConfigStruct import Parameter, Cumulative_Integer_Struct, LTP_Parameter 

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""

def init_config():

  cs = CS.ConfigurationSpace()

  conv_ops = ["StdConv", "Conv3", "Conv5","Conv7","MaxPool5","AvgPool5","MaxPool7","AvgPool7","SepConv3","SepConv5","SepConv7", "Identity"]
  
  ops_parameters = [
        Parameter("type",               "Categorical", lower_or_constant_value = conv_ops ), 
        LTP_Parameter("input_1",               "Integer", 0,10),
        LTP_Parameter("input_2",               "Integer", 0,10)]
  ops = Cumulative_Integer_Struct(cs,ops_parameters,"ops","num_ops","Integer",1,9)


  conv_parameters = [
        ops]
 
  Cumulative_Integer_Struct(cs,conv_parameters,"normal_cell", "num_conv","Integer", 1, 1).init() 

  conv_ops = ["FactorizedReduce"]
  
  ops_parameters = [
        Parameter("type",               "Categorical", lower_or_constant_value = conv_ops ), 
        LTP_Parameter("input_1",               "Integer", 0,10),
        LTP_Parameter("input_2",               "Integer", 0,10)]
  ops = Cumulative_Integer_Struct(cs,ops_parameters,"ops","num_ops","Integer",1,1)


  conv_parameters = [
        ops]
 
  Cumulative_Integer_Struct(cs,conv_parameters,"reduction_cell", "num_re","Integer", 1, 1).init() 
  """
  ops_type_list = ["StdConv"]
  
  ops_parameters = [
        Parameter("type",               "Categorical", lower_or_constant_value = ops_type_list ), 
        LTP_Parameter("input",               "Integer", 0,10)]
  ops = Cumulative_Integer_Struct(cs,ops_parameters,"ops","num_ops","Integer",1,5)


  conv_parameters = [
        Parameter("type",               "Constant", lower_or_constant_value = "Conv1D"),
        Parameter("padding",            "Constant" ,lower_or_constant_value = "same"),
        Parameter("filters",            "Constant", lower_or_constant_value =  1),
        Parameter("BatchNormalization", "Integer", 0,1),
        Parameter("kernel_size",        "Integer", 1,16),
        Parameter("kernel_regularizer", "Float", 1e-8,5e-1, log = True),
        Parameter("bias_regularizer",   "Float", 1e-8,5e-1, log = True),
        Parameter("activity_regularizer", "Float", 1e-8,5e-1, log = True),
        ops]
 
  Cumulative_Integer_Struct(cs,conv_parameters,"cell", "num_cells","Integer", 1, 5).init() 


    
  dense_parameters = [
        Parameter("type",               "Constant", "Dense"),
        Parameter("units",              "Integer", 1,128),
        Parameter("kernel_regularizer", "Float", 1e-8,5e-1, log = True),
        Parameter("bias_regularizer",   "Float", 1e-8,5e-1, log = True),
     self.train_dataset = Train_BTC()
       Parameter("activity_regularizer", "Float", 1e-8,5e-1, log = True)]
     
  Cumulative_Integer_Struct(cs,dense_parameters,"dense","num_dense_layers","Integer", 1, 3).init() 
  """
    ###Training Configuration###
    ###Optimiser###
  lr =CSH.UniformFloatHyperparameter(name = "lr",			lower = 0.000001  ,upper = 0.05)
  p =CSH.UniformFloatHyperparameter(name = "p",			lower = 0.01 ,upper = 0.3 )
  epochs = CSH.UniformIntegerHyperparameter(name = "epochs", lower = 1, upper = 50)
  layers = CSH.UniformIntegerHyperparameter(name = "layers", lower = 1, upper = 6)
  c1 = CSH.UniformFloatHyperparameter(name = "c1_weight" , lower = 1,upper = 3)
  batch_size = CSH.UniformIntegerHyperparameter(name = "batch_size", lower = 2, upper = 16)
  T_0 = CSH.UniformIntegerHyperparameter(name = "T_0", lower = 2, upper = 10)
  T_mult = CSH.UniformIntegerHyperparameter(name = "T_mult", lower = 1, upper = 5)
  channels = CSH.UniformIntegerHyperparameter(name = "channels", lower = 2, upper = 30)
  augmentations = CSH.UniformIntegerHyperparameter(name = "augmentations", lower = 0, upper = 20)
  
    ###Topology Definition]###
  
  hp_list = [
        c1,
        epochs,
        lr,
        p,
        layers,
        T_0,
        T_mult,
        batch_size,
        channels,
        augmentations
 
]
  cs.add_hyperparameters(hp_list)
  return cs

if __name__ == "__main__":
	configS = init_config()	
	print(configS.get_hyperparameters())



import os
import numpy as np
import torch
import shutil
from torch.autograd import Variable
import json
from sklearn.metrics import confusion_matrix


class AvgrageMeter(object):

  def __init__(self):
    self.reset()

  def reset(self):
    self.avg = 0
    self.sum = 0
    self.cnt = 0

  def update(self, val, n=1):
    self.sum += val * n
    self.cnt += n
    self.avg = self.sum / self.cnt


def accuracy(output, target, topk=(1,)):
  maxk = max(topk)
  batch_size = target.size(0)

  _, pred = output.topk(maxk, 1, True, True)
  pred = pred.t()
  correct = pred.eq(target.view(1, -1).expand_as(pred))
  print(correct.shape)
  res = []
  for k in topk:
    correct_k = correct[:k].reshape(-1).float().sum(0)
    res.append(correct_k.mul_(100.0/batch_size))
  return res


class Cutout(object):
    def __init__(self, length):
        self.length = length

    def __call__(self, img):
        h, w = img.size(1), img.size(2)
        mask = np.ones((h, w), np.float32)
        y = np.random.randint(h)
        x = np.random.randint(w)

        y1 = np.clip(y - self.length // 2, 0, h)
        y2 = np.clip(y + self.length // 2, 0, h)
        x1 = np.clip(x - self.length // 2, 0, w)
        x2 = np.clip(x + self.length // 2, 0, w)

        mask[y1: y2, x1: x2] = 0.
        mask = torch.from_numpy(mask)
        mask = mask.expand_as(img)
        img *= mask
        return img


def _data_transforms_cifar10(args):
  CIFAR_MEAN = [0.49139968, 0.48215827, 0.44653124]
  CIFAR_STD = [0.24703233, 0.24348505, 0.26158768]

  train_transform = transforms.Compose([
    transforms.RandomCrop(32, padding=4),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
  ])
  if args.cutout:
    train_transform.transforms.append(Cutout(args.cutout_length))

  valid_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
    ])
  return train_transform, valid_transform


def count_parameters_in_MB(model):
  return np.sum(np.prod(v.size()) for name, v in model.named_parameters() if "auxiliary" not in name)/1e6


def save_checkpoint(state, is_best, save):
  filename = os.path.join(save, 'checkpoint.pth.tar')
  torch.save(state, filename)
  if is_best:
    best_filename = os.path.join(save, 'model_best.pth.tar')
    shutil.copyfile(filename, best_filename)


def save(model, model_path):
  torch.save(model.state_dict(), model_path)


def load(model, model_path):
  model.load_state_dict(torch.load(model_path))

"""
def drop_path(x, drop_prob):
  if drop_prob > 0.:
    keep_prob = 1.-drop_prob
    mask = Variable(torch.cuda.FloatTensor(x.size(0),1, 1).bernoulli_(keep_prob))
    x.div_(keep_prob)
    x.mul_(mask)
  return x
"""
def drop_path2d(x, drop_prob):
  if drop_prob > 0.:
    keep_prob = 1.-drop_prob
    mask = Variable(torch.cuda.FloatTensor(x.size(0), 1,1, 1).bernoulli_(keep_prob))
    x.div_(keep_prob)
    x.mul_(mask)
  return x

def drop_path(x, drop_prob):
  if drop_prob > 0.:
    keep_prob = 1.-drop_prob
    mask = Variable(torch.cuda.FloatTensor(x.size(0), 1, 1).bernoulli_(keep_prob))
    x.div_(keep_prob)
    x.mul_(mask)
  return x

def create_exp_dir(path, scripts_to_save=None):
  if not os.path.exists(path):
    os.mkdir(path)
  print('Experiment dir : {}'.format(path))

  if scripts_to_save is not None:
    os.mkdir(os.path.join(path, 'scripts'))
    for script in scripts_to_save:
      dst_file = os.path.join(path, 'scripts', os.path.basename(script))
      shutil.copyfile(script, dst_file)



import pandas as pd


class BernoulliLogger:
  def __init__(self,PATH,ID): 
    if not os.path.exists("{}/{}".format(PATH,"metrics")):
      os.mkdir("{}/metrics".format(self.PATH))
    self.path = "{}/{}/{}-bin.npy".format(PATH,"metrics",ID)
  def update(self,data):
    if os.path.exists(self.path):
      x = np.load(self.path)
      np.save(self.path,np.vstack((x,data)))
    else:
      np.save(self.path,data)

class MetricLogger:
  def __init__(self,PATH):
    self.performance_data = {"ID": [],"accuracy": [], "recall":[]}
    self.PATH = PATH
    if not os.path.exists("{}/{}".format(PATH,"metrics")):
      os.mkdir("{}/metrics".format(self.PATH))
    self.initial = True 
  def load(self,ID):
    path = "{}/{}/{}".format(self.PATH,"metrics",ID)
    data = pd.read_csv(path).to_dict('list')
    self.performance_data = {"ID": data["ID"],"accuracy": data["accuracy"], "recall":data["recall"]}

  def update(self, new_data : dict):
    if self.initial and os.path.exists("{}/{}/{}".format(self.PATH,"metrics",new_data["ID"])):
      self.load(new_data["ID"])
      self.initial = False
    for i in new_data:
      self.performance_data[i].append(new_data[i])

    df = pd.DataFrame.from_dict(self.performance_data)
    df.to_csv("{}/{}/{}".format(self.PATH,"metrics",new_data["ID"]))




def calculate_train_vals(pred_tensor,gt_tensor):
    # Convert the tensors to NumPy arrays
  predictions = torch.argmax(pred_tensor, dim=1).numpy()
  labels = gt_tensor.numpy()
  # Calculate accuracy
  correct = np.sum(predictions == labels)
  total = len(labels)
  accuracy = correct / total

  # Determine the number of classes
  num_classes = int(max(labels.max(), predictions.max()) + 1)

  # Generate confusion matrix
  cm = confusion_matrix(labels, predictions)

  # Serialize the confusion matrix as a JSON string
  confusion_matrix_str = json.dumps(cm.tolist())
  return confusion_matrix_str, accuracy






import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""
def fully_connected_cell( num_nodes : int, num_operations : int , operation_list : list ): 
  cs = CS.ConfigurationSpace()

  for out_node in range(1,num_nodes):
    for in_node in range(0,out_node):
      op = CSH.CategoricalHyperparameter(name =  str(in_node)+"-"+str(out_node), choices = operation_list)
      cs.add_hyperparameter(op)
  return cs

def nats_bench_topology( V : int  = 4, L : int = 5):
  OPS = ["nor_conv_3x3", "nor_conv_1x1","none" , "avg_pool_3x3", "skip_connect"]
  return fully_connected_cell(V,L,OPS)

if __name__ == "__main__":
  cs = nats_bench_topology()
  print(cs.sample_configuration())   



import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH
import copy

"""
Christopher MacKinnon 2021

The objective of this class is a layer of abstraction on top of the ConfigSpace package, allowing
parameters to be defined in a hierarchy rather than absolutely defining each parameter. This is 
aimed at allowing the hyperparameter for each layer in a model to be dynamically constructed once 
a layer is defined 
    
"""

class Parameter:
    """
    Base parameter class. This is effectively a wrapper around the ConfigSpace hyperparameters
    """ 
    def __init__(self, name : str , hyperparameter_type : str, lower_or_constant_value , upper_value = None ,normal = False, log : bool = False):
        self.name = name
        self.type = hyperparameter_type
        self.lower = lower_or_constant_value 
        self.upper = upper_value
        if self.lower == self.upper:
          self.type = "Constant"
          print("Warning: Set type as integer but range same upper and lower values given, Setting to Constant")
        self.normal = normal 
        self.log = log 
        if self.type == "Constant":
            self.config = CSH.Constant(name = name, value = lower_or_constant_value)
        if normal == False:
            if self.type == "Integer":
                self.config = CSH.UniformIntegerHyperparameter(name = name, lower = lower_or_constant_value, upper = upper_value, log = log)

            if self.type == "Float":
                self.config = CSH.UniformFloatHyperparameter(name = name, lower = lower_or_constant_value, upper = upper_value, log = log)
         
        if normal == True:
            if self.type == "Integer":
                self.config = CSH.NormalIntegerHyperparameter(name = name, lower = lower_or_constant_value, upper = upper_value, log = log)

            if self.type == "Float":
                self.config = CSH.NormalFloatHyperparameter(name = name, lower = lower_or_constant_value, upper = upper_value, log = log)
        if self.type == "Categorical":
          self.config = CSH.CategoricalHyperparameter(name = name , choices = lower_or_constant_value)

    def has_children(self):
        return False

    def has_unexpanded_children(self):
        return self.has_children()

    def get_args(self, args = None):
    
        return self.type, self.lower, self.upper, self.normal, self.log
    
    def set_arg_upper(self, value):
        self.upper = value

class LTP_Parameter(Parameter):
    #Less Than Parent, Sets upper limit to value of parent from string (i.e. conv_5_size.upper = 4)
  def __init__(self, name : str , hyperparameter_type : str, lower_or_constant_value , upper_value = None ,normal = False, log : bool = False):
    super().__init__( name , hyperparameter_type , lower_or_constant_value , upper_value ,normal , log )

  def get_args(self, parent_string = None):
    if parent_string != None:
      self.upper = int(parent_string[:-(len("_"+self.name))][-1]) - 1 
    return self.type, self.lower, self.upper, self.normal, self.log


class Integer_Struct(Parameter):
    """
    Generates 1 of each child parameter for each integer between lower and upper and adds conditions 
    name of children follows format parent_name + _ + number + _ + child_name
    """
    
    def __init__(self, config_space,  children : list,struct_name : str , name : str , hyperparameter_type : str, lower_or_constant_value , upper_value = None ,normal = False, log :  bool = False ):
        
        super().__init__(name, hyperparameter_type, lower_or_constant_value, upper_value, normal, log)
        self.children_template = children
        self.children = []
        self.children_dict = dict()
        self.conditions = []
        self.config_space = config_space
        self.struct_name = struct_name 
        self.expanded = False
    def add_children_to_config_space(self):
        self.config_space.add_hyperparameters( [child.config for child in self.children] )
        self.config_space.add_hyperparameter(self.config)

    def generate_name(self,parent_name : str, child_name : str, num : int):
        return parent_name + "_" + str(num) +"_" + child_name  

    def has_unexpanded_children(self):
      if self.expanded == False:
        self.expanded = True 
        return True    
      else:
        return False

    def expand_children(self):
        for i in self.children_template:
          if i.has_unexpanded_children():
            p_children = i.generate_children()
            self.children_template.extend(p_children)
            


    def _generate_child_set(self, i ):
        #Generates children for iteration i of parent
        children = list()
        
        for child_template in self.children_template:
            new_child_name = self.generate_name(self.struct_name, child_template.name , i )
            print(new_child_name , child_template.get_args())
            children.append( Parameter( new_child_name, *child_template.get_args(new_child_name) ))
        return children

    def _generate_conditions(self):
        pass

    def generate_children(self, rec = False):
        #Generate child parameter for each value of the parent class
        for i in range(self.lower, self.upper + 1): 
            children = self._generate_child_set(i)
            self.children += children
            self.children_dict[i] = children
        return self.children

class Cumulative_Integer_Struct(Integer_Struct):
    """ (CIS)
    Generates an instance for each child parameter for 1 to n (i.e. 1,2,..,n ) where the parent parameter 
    value is n.
    """
    def __init__(self,config_space ,  children : list,struct_name, name : str ,
                hyperparameter_type : str, lower_or_constant_value , 
                upper_value = None ,normal = False, log :bool = False ):

        super().__init__(config_space ,children, struct_name ,name, 
                hyperparameter_type, lower_or_constant_value, upper_value, normal, log)



     
    def batch_add_greater_than_cond(self, a_list, b , num):
        for a in a_list:
            if num != self.lower: 
                cond = CS.GreaterThanCondition(a.config,b,num -1 )
                self.config_space.add_condition(cond)  
    
    def generate_conditions(self):
        for i in range(self.lower, self.upper + 1):
            self.batch_add_greater_than_cond( self.children_dict[i] , self.config , i )
    def init(self):
        self.expand_children()
        self.generate_children()
        self.add_children_to_config_space()
        self.generate_conditions()



if __name__ == "__main__":
    pass        



# Author: Hassan Ismail Fawaz <hassan.ismail-fawaz@uha.fr>
#         Germain Forestier <germain.forestier@uha.fr>
#         Jonathan Weber <jonathan.weber@uha.fr>
#         Lhassane Idoumghar <lhassane.idoumghar@uha.fr>
#         Pierre-Alain Muller <pierre-alain.muller@uha.fr>
# License: GPL3

import numpy as np
import pandas as pd
import matplotlib

matplotlib.use('agg')
import matplotlib.pyplot as plt

matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = 'Arial'

import operator
import math
from scipy.stats import wilcoxon
from scipy.stats import friedmanchisquare
import networkx

# inspired from orange3 https://docs.orange.biolab.si/3/data-mining-library/reference/evaluation.cd.html
def graph_ranks(avranks, names, p_values, cd=None, cdmethod=None, lowv=None, highv=None,
                width=6, textspace=1, reverse=False, filename=None, labels=False, **kwargs):
    """
    Draws a CD graph, which is used to display  the differences in methods'
    performance. See Janez Demsar, Statistical Comparisons of Classifiers over
    Multiple Data Sets, 7(Jan):1--30, 2006.

    Needs matplotlib to work.

    The image is ploted on `plt` imported using
    `import matplotlib.pyplot as plt`.

    Args:
        avranks (list of float): average ranks of methods.
        names (list of str): names of methods.
        cd (float): Critical difference used for statistically significance of
            difference between methods.
        cdmethod (int, optional): the method that is compared with other methods
            If omitted, show pairwise comparison of methods
        lowv (int, optional): the lowest shown rank
        highv (int, optional): the highest shown rank
        width (int, optional): default width in inches (default: 6)
        textspace (int, optional): space on figure sides (in inches) for the
            method names (default: 1)
        reverse (bool, optional):  if set to `True`, the lowest rank is on the
            right (default: `False`)
        filename (str, optional): output file name (with extension). If not
            given, the function does not write a file.
        labels (bool, optional): if set to `True`, the calculated avg rank
        values will be displayed
    """
    try:
        import matplotlib
        import matplotlib.pyplot as plt
        from matplotlib.backends.backend_agg import FigureCanvasAgg
    except ImportError:
        raise ImportError("Function graph_ranks requires matplotlib.")

    width = float(width)
    textspace = float(textspace)

    def nth(l, n):
        """
        Returns only nth elemnt in a list.
        """
        n = lloc(l, n)
        return [a[n] for a in l]

    def lloc(l, n):
        """
        List location in list of list structure.
        Enable the use of negative locations:
        -1 is the last element, -2 second last...
        """
        if n < 0:
            return len(l[0]) + n
        else:
            return n

    def mxrange(lr):
        """
        Multiple xranges. Can be used to traverse matrices.
        This function is very slow due to unknown number of
        parameters.

        >>> mxrange([3,5])
        [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]

        >>> mxrange([[3,5,1],[9,0,-3]])
        [(3, 9), (3, 6), (3, 3), (4, 9), (4, 6), (4, 3)]

        """
        if not len(lr):
            yield ()
        else:
            # it can work with single numbers
            index = lr[0]
            if isinstance(index, int):
                index = [index]
            for a in range(*index):
                for b in mxrange(lr[1:]):
                    yield tuple([a] + list(b))

    def print_figure(fig, *args, **kwargs):
        canvas = FigureCanvasAgg(fig)
        canvas.print_figure(*args, **kwargs)

    sums = avranks

    nnames = names
    ssums = sums

    if lowv is None:
        lowv = min(1, int(math.floor(min(ssums))))
    if highv is None:
        highv = max(len(avranks), int(math.ceil(max(ssums))))

    cline = 0.4

    k = len(sums)

    lines = None

    linesblank = 0
    scalewidth = width - 2 * textspace

    def rankpos(rank):
        if not reverse:
            a = rank - lowv
        else:
            a = highv - rank
        return textspace + scalewidth / (highv - lowv) * a

    distanceh = 0.25

    cline += distanceh

    # calculate height needed height of an image
    minnotsignificant = max(2 * 0.2, linesblank)
    height = cline + ((k + 1) / 2) * 0.2 + minnotsignificant

    fig = plt.figure(figsize=(width, height))
    fig.set_facecolor('white')
    ax = fig.add_axes([0, 0, 1, 1])  # reverse y axis
    ax.set_axis_off()

    hf = 1. / height  # height factor
    wf = 1. / width

    def hfl(l):
        return [a * hf for a in l]

    def wfl(l):
        return [a * wf for a in l]

    # Upper left corner is (0,0).
    ax.plot([0, 1], [0, 1], c="w")
    ax.set_xlim(0, 1)
    ax.set_ylim(1, 0)

    def line(l, color='k', **kwargs):
        """
        Input is a list of pairs of points.
        """
        ax.plot(wfl(nth(l, 0)), hfl(nth(l, 1)), color=color, **kwargs)

    def text(x, y, s, *args, **kwargs):
        ax.text(wf * x, hf * y, s, *args, **kwargs)

    line([(textspace, cline), (width - textspace, cline)], linewidth=2)

    bigtick = 0.3
    smalltick = 0.15
    linewidth = 2.0
    linewidth_sign = 4.0

    tick = None
    for a in list(np.arange(lowv, highv, 0.5)) + [highv]:
        tick = smalltick
        if a == int(a):
            tick = bigtick
        line([(rankpos(a), cline - tick / 2),
              (rankpos(a), cline)],
             linewidth=2)

    for a in range(lowv, highv + 1):
        text(rankpos(a), cline - tick / 2 - 0.05, str(a),
             ha="center", va="bottom", size=16)

    k = len(ssums)

    def filter_names(name):
        return name

    space_between_names = 0.24

    for i in range(math.ceil(k / 2)):
        chei = cline + minnotsignificant + i * space_between_names
        line([(rankpos(ssums[i]), cline),
              (rankpos(ssums[i]), chei),
              (textspace - 0.1, chei)],
             linewidth=linewidth)
        if labels:
            text(textspace + 0.3, chei - 0.075, format(ssums[i], '.4f'), ha="right", va="center", size=10)
        text(textspace - 0.2, chei, filter_names(nnames[i]), ha="right", va="center", size=16)

    for i in range(math.ceil(k / 2), k):
        chei = cline + minnotsignificant + (k - i - 1) * space_between_names
        line([(rankpos(ssums[i]), cline),
              (rankpos(ssums[i]), chei),
              (textspace + scalewidth + 0.1, chei)],
             linewidth=linewidth)
        if labels:
            text(textspace + scalewidth - 0.3, chei - 0.075, format(ssums[i], '.4f'), ha="left", va="center", size=10)
        text(textspace + scalewidth + 0.2, chei, filter_names(nnames[i]),
             ha="left", va="center", size=16)

    # no-significance lines
    def draw_lines(lines, side=0.05, height=0.1):
        start = cline + 0.2

        for l, r in lines:
            line([(rankpos(ssums[l]) - side, start),
                  (rankpos(ssums[r]) + side, start)],
                 linewidth=linewidth_sign)
            start += height
            print('drawing: ', l, r)

    # draw_lines(lines)
    start = cline + 0.2
    side = -0.02
    height = 0.1

    # draw no significant lines
    # get the cliques
    cliques = form_cliques(p_values, nnames)
    i = 1
    achieved_half = False
    print(nnames)
    for clq in cliques:
        if len(clq) == 1:
            continue
        print(clq)
        min_idx = np.array(clq).min()
        max_idx = np.array(clq).max()
        if min_idx >= len(nnames) / 2 and achieved_half == False:
            start = cline + 0.25
            achieved_half = True
        line([(rankpos(ssums[min_idx]) - side, start),
              (rankpos(ssums[max_idx]) + side, start)],
             linewidth=linewidth_sign)
        start += height


def form_cliques(p_values, nnames):
    """
    This method forms the cliques
    """
    # first form the numpy matrix data
    m = len(nnames)
    g_data = np.zeros((m, m), dtype=np.int64)
    for p in p_values:
        if p[3] == False:
            i = np.where(nnames == p[0])[0][0]
            j = np.where(nnames == p[1])[0][0]
            min_i = min(i, j)
            max_j = max(i, j)
            g_data[min_i, max_j] = 1

    g = networkx.Graph(g_data)
    return networkx.find_cliques(g)


def draw_cd_diagram(df_perf=None, alpha=0.05, title=None, labels=False):
    """
    Draws the critical difference diagram given the list of pairwise classifiers that are
    significant or not
    """
    p_values, average_ranks, _ = wilcoxon_holm(df_perf=df_perf, alpha=alpha)

    print(average_ranks)

    for p in p_values:
        print(p)


    graph_ranks(average_ranks.values, average_ranks.keys(), p_values,
                cd=None, reverse=True, width=9, textspace=1.5, labels=labels)

    font = {'family': 'sans-serif',
        'color':  'black',
        'weight': 'normal',
        'size': 22,
        }
    if title:
        plt.title(title,fontdict=font, y=0.9, x=0.5)
    plt.savefig('cd-diagram.png',bbox_inches='tight')

def wilcoxon_holm(alpha=0.05, df_perf=None):
    """
    Applies the wilcoxon signed rank test between each pair of algorithm and then use Holm
    to reject the null's hypothesis
    """
    print(pd.unique(df_perf['classifier_name']))
    # count the number of tested datasets per classifier
    df_counts = pd.DataFrame({'count': df_perf.groupby(
        ['classifier_name']).size()}).reset_index()
    # get the maximum number of tested datasets
    max_nb_datasets = df_counts['count'].max()
    # get the list of classifiers who have been tested on nb_max_datasets
    classifiers = list(df_counts.loc[df_counts['count'] == max_nb_datasets]
                       ['classifier_name'])
    # test the null hypothesis using friedman before doing a post-hoc analysis
    friedman_p_value = friedmanchisquare(*(
        np.array(df_perf.loc[df_perf['classifier_name'] == c]['accuracy'])
        for c in classifiers))[1]
    if friedman_p_value >= alpha:
        # then the null hypothesis over the entire classifiers cannot be rejected
        print('the null hypothesis over the entire classifiers cannot be rejected')
        exit()
    # get the number of classifiers
    m = len(classifiers)
    # init array that contains the p-values calculated by the Wilcoxon signed rank test
    p_values = []
    # loop through the algorithms to compare pairwise
    for i in range(m - 1):
        # get the name of classifier one
        classifier_1 = classifiers[i]
        # get the performance of classifier one
        perf_1 = np.array(df_perf.loc[df_perf['classifier_name'] == classifier_1]['accuracy']
                          , dtype=np.float64)
        for j in range(i + 1, m):
            # get the name of the second classifier
            classifier_2 = classifiers[j]
            # get the performance of classifier one
            perf_2 = np.array(df_perf.loc[df_perf['classifier_name'] == classifier_2]
                              ['accuracy'], dtype=np.float64)
            # calculate the p_value
            p_value = wilcoxon(perf_1, perf_2, zero_method='pratt')[1]
            # appen to the list
            p_values.append((classifier_1, classifier_2, p_value, False))
    # get the number of hypothesis
    k = len(p_values)
    # sort the list in acsending manner of p-value
    p_values.sort(key=operator.itemgetter(2))

    # loop through the hypothesis
    for i in range(k):
        # correct alpha with holm
        new_alpha = float(alpha / (k - i))
        # test if significant after holm's correction of alpha
        if p_values[i][2] <= new_alpha:
            p_values[i] = (p_values[i][0], p_values[i][1], p_values[i][2], True)
        else:
            # stop
            break
    # compute the average ranks to be returned (useful for drawing the cd diagram)
    # sort the dataframe of performances
    sorted_df_perf = df_perf.loc[df_perf['classifier_name'].isin(classifiers)]. \
        sort_values(['classifier_name', 'dataset_name'])
    # get the rank data
    rank_data = np.array(sorted_df_perf['accuracy']).reshape(m, max_nb_datasets)

    # create the data frame containg the accuracies
    df_ranks = pd.DataFrame(data=rank_data, index=np.sort(classifiers), columns=
    np.unique(sorted_df_perf['dataset_name']))

    # number of wins
    dfff = df_ranks.rank(ascending=False)
    print(dfff[dfff == 1.0].sum(axis=1))

    # average the ranks
    average_ranks = df_ranks.rank(ascending=False).mean(axis=1).sort_values(ascending=False)
    # return the p-values and the average ranks
    return p_values, average_ranks, max_nb_datasets






import random


class SuperSet:
  def __init__(self, dataloaders):
    self.dataloaders = dataloaders
    self.iterators = []
    self.set_iterator()
  def __getitem__(self, index):
    rand_index = self.get_idx()
    try:
      return next(self.iterators[rand_index])
    except StopIteration:
      self.iterators.pop(rand_index)
    if len(self.iterators) == 0:
      raise StopIteration
    rand_index = self.get_idx()
    return next(self.iterators[rand_index])

  def set_iterator(self):
    for i in self.dataloaders:
      self.iterators.append(iter(i))

  def get_idx(self):
    if len(self.iterators) == 1:
      return 0
    elif len(self.iterators) == 0:
      self.set_iterator()
      raise StopIteration


    else:
      return random.randint(0,len(self.iterators) -1)

  def __len__(self):
    return sum([len(x) for x in self.dataloaders])  













import numpy as np 
import pandas as pd
import os

import os
import sqlite3
import time
class Logger:
    def __init__(self, path, experiment_name,dataset,fold,repeat,parameters):

        self.fold = fold
        self.parameters = parameters
        self.repeat = repeat
        self.experiment = experiment_name
        self.dataset = dataset
        # Set up the database connection
        self.conn = sqlite3.connect(path)
        self.c = self.conn.cursor()


    def update(self, data_dict):
        start_time = time.time()
        timeout=5
        retry_interval=0.05
        while time.time() - start_time < timeout:
            try:
                # Insert the provided data into the training_info table
                self.c.execute("""
                INSERT OR REPLACE INTO training_info (experiment, dataset,  model_id, epoch, training_loss, validation_loss, training_accuracy, validation_accuracy, learning_rate, confusion_matrix_train,confusion_matrix_test,fold,repeat,parameters)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ? ,? ,? ,? )
                """, (self.experiment, 
                  self.dataset, 
                  data_dict["ID"], 
                  data_dict["epoch"], 
                  data_dict["loss"], 
                  data_dict["validation_loss"], 
                  data_dict["training_accuracy"], 
                  data_dict["validation_accuracy"], 
                  data_dict["lr"],
                  data_dict["confusion_matrix_train"],
                  data_dict["confusion_matrix_test"],
                self.fold, 
                self.repeat,
                self.parameters
                )
                )
                self.conn.commit()
            except sqlite3.OperationalError as e:
                if "database is locked" in str(e):
                    # Wait for the retry interval before attempting again
                    time.sleep(retry_interval)
                    continue
                else:
                    # Raise the exception if it's not related to a locked database
                    raise e


    def close(self):
        # Close the database connection when done
        self.conn.close()


import numpy as np
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
from HPO.utils.time_series_augmentation import permutation , magnitude_warp, time_warp, window_slice, jitter, scaling, rotation
import os
def none(x):
  return x
class TEPS(Dataset):
  def __init__(self, window_size, x : str , y : str ):
    path = "datasets/TEPS/"
    self.x = torch.from_numpy(np.reshape(np.load(path+x),(-1,52)))
    self.y = torch.from_numpy(np.load(path+y))
    
    self.n_samples = self.x.shape[0]
    self.window = window_size
    self.n_classes = len(np.unique(self.y))

  def set_window_size(self, window_size):
    self.window = window_size
  def __getitem__(self, index):
    while index+self.window > self.n_samples:
      index = random.randint(0,self.n_samples)
    x = self.x[index:index+self.window]
    y = self.y[index+self.window-1]
    x = x.reshape(52,self.window)
    return x, y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples - self.n_samples%self.window


class Train_TEPS(TEPS):

  def __init__(self, window_size = 200): 
    super().__init__(window_size, "x_train.npy","y_train.npy")

class Test_TEPS(TEPS):
  def __init__(self, window_size = 200): 
    super().__init__(window_size, "x_test.npy","y_test.npy")

class BTC(Dataset):
  def __init__(self, window_size, prediction_distance, split):
    self.path ="{}/scripts/datasets/BTC/BTC.npy".format(os.environ["HOME"])
    self.prediction_distance = prediction_distance
    self.x = torch.from_numpy(np.reshape(np.load(self.path),(-1,10)))[1000000:,:]
    if split > 0.5:
      self.x = self.x[:int(self.x.shape[0]*split)]
    else:
      self.x = self.x[-int(self.x.shape[0]*split):]
    self.n_samples = self.x.shape[0]
    ss = StandardScaler()
    self.x = ss.fit_transform(self.x)
    self.window = window_size
    self.generate_percent_labels()
    self.n_features = self.x.shape[1]
    self.n_classes = len(np.unique(self.y))
    print(self.y)
    for i in range(self.n_classes):
      print(np.count_nonzero(self.y == i))

    print(len(np.unique(self.y)))
  def generate_labels(self):
    idx = 3
    self.y = []
    for i in range(self.n_samples- self.prediction_distance):
        if self.x[i+self.prediction_distance,idx] > self.x[i,idx]:
          self.y.append(1)
        else:
          self.y.append(0)
    print(self.x.shape[0])


  def generate_percent_labels(self, classes = [0.01,0.1,0.25,0.5,0.75,0.9,0.99,1]):
    abs_change = np.diff(self.x[:,3])
    self.class_labels = {}
    per_change = np.divide(abs_change , self.x[1:,3])
    i_last = 0 
    self.y = np.zeros(self.x.shape[0])
    for c,i in enumerate(classes):
      self.class_labels[c] = "{}-{}".format(i_last,i)
      print(np.nanquantile(per_change, i))
      hold = np.where( (per_change > np.nanquantile(per_change,i_last)) & (per_change < np.nanquantile(per_change,i)))
      i_last = i 
      self.y[hold] = c
    
    print(len(self.y))
    print(self.x.shape[0])



  def set_window_size(self, window_size):
    self.window = window_size
  def __getitem__(self, index):
    while index+self.window+self.prediction_distance > self.n_samples:
      index = random.randint(0,self.n_samples)
    x = self.x[index:index+self.window]
    y = self.y[index+self.window-1]
    x = x.reshape(10,self.window)
    return x, y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_features

  def __len__(self):
    return self.n_samples - self.n_samples%self.window


class Train_BTC(BTC):

  def __init__(self, window_size = 100, split = 0.95 , pred_dist = 1 ): 
    super().__init__(window_size, split = split, prediction_distance = pred_dist)

class Test_BTC(BTC):
  def __init__(self, window_size = 100, split = 0.05, pred_dist = 1): 
    super().__init__(window_size,split = split, prediction_distance = pred_dist)



class repsol(Dataset):
  def __init__(self, window_size, files : list , augmentations):
    path = "/home/snaags/scripts/datasets/"
    data = []
    self.x_index_address = {}
    self.y_index_address = {}
    self.window = window_size
    augmentations_per_batch = 2
    #index_list sum()
    if augmentations == True:
      self.non_warp_augmentations =[jitter, scaling, rotation, permutation ]
      self.warp_augmentations = [magnitude_warp, time_warp]
    for i in files:
      data.append(np.reshape(np.load(path+i),(-1,28)))
    self.current_index = 0
    use_all = True

  
    for batch in data:
      batch_x = batch[:,1:]
      batch_y = batch[:,0]
      self.add_to_dataset(batch_x,batch_y)
      if augmentations:
        if use_all == True:
          x = batch_x.reshape(1,batch_x.shape[0],batch_x.shape[1])
          augs = self.warp_augmentations +self.non_warp_augmentations
          for i in range(len(augs)):
              x = augs.pop(random.randint(0,len(augs) -1))(x)
          self.add_to_dataset(x.reshape(*x.shape[1:]),batch_y)
        else:
          for i in range(augmentations_per_batch):
            augmentations_for_current_batch = []
            temp_augmentation_list = self.non_warp_augmentations.copy()
            augmentations_for_current_batch.append(random.choice(self.warp_augmentations))
            x = batch_x.reshape(1,batch_x.shape[0],batch_x.shape[1])
            while len(augmentations_for_current_batch) < 3:
              augmentations_for_current_batch.append(temp_augmentation_list.pop(random.randint(0,len(temp_augmentation_list)-1)))
      
            for augmentation_function in augmentations_for_current_batch:
              x = augmentation_function(x)
            self.add_to_dataset(x.reshape(*x.shape[1:]),batch_y)

    
    self.n_samples = self.current_index #* len(self.augmentations)
    self.features = 27
    self.n_classes = 2
  def add_to_dataset(self,x,y):
    self.x_index_address[self.current_index] = torch.from_numpy(x)
    self.y_index_address[self.current_index] = torch.from_numpy(y)
    self.current_index += x.shape[0] - self.window

  def set_window_size(self, window_size):
    self.window = window_size

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index
    addr = 0
    augmentation_idx = 0
    #while index > self.real_samples:
    #  index -= self.real_samples
    #  augmentation_idx += 1 


    for i in self.x_index_address:
      if (index - i) < 0:
        break
      else:
        addr = i

    index -= addr 
    x = self.x_index_address[addr]
    y = self.y_index_address[addr]

    x = x[index:index+self.window]
    y = y[index+self.window-1]

    return x.reshape( 27,self.window ) , y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples - self.n_samples%self.window


class Train_repsol(repsol):

  def __init__(self, train_files, window_size = 200, augmentations = True): 
    super().__init__(window_size, train_files, augmentations)

class Test_repsol(repsol):

  def __init__(self,test_files, window_size = 200, augmentations = False): 
    super().__init__(window_size, test_files , augmentations)


class TEPS_split(Dataset):
  def __init__(self, window_size, train , augmentations, max_samples):


    self.features = 52
    self.n_classes = 21
    self.samples_per_class = [0]* self.n_classes
    self.max_samples = max_samples

    path = "/home/snaags/scripts/datasets/TEPS/split/"
    files = os.listdir(path)
    if train == True:
      filtr = "training"
    else:
      filtr = "testing"
    files = [ name for name in files if filtr in name]    
    
    num_files = max_samples / 500 
    while len(files) > num_files:
      files.pop(random.randint(0, len(files)-1))
     
    data = []
    self.x_index_address = {}
    self.y_index_address = {}
    self.window = window_size
    while len(files):
      data.append(np.reshape(np.load(path+files.pop(random.randint(0, len(files)-1))),(-1,53)))
    self.current_index = 0
    use_all = True
  
    for batch in data:
      batch_x = batch[:,1:]
      batch_y = batch[:,0]
      self.add_to_dataset(batch_x,batch_y)

    self.n_samples = self.current_index #* len(self.augmentations)
 
  def add_to_dataset(self,x,y):
    self.x_index_address[self.current_index] = torch.from_numpy(x.reshape(52,-1)).float()
    self.y_index_address[self.current_index] = torch.from_numpy(np.unique(y)).long()
    self.samples_per_class[int(y[0])] += x.shape[0]
    self.current_index += 1

  def get_n_samples_per_class(self):
    for i in range(self.n_classes):
      print("Samples in class {}: {}".format(i, self.samples_per_class[i]))    

  def set_window_size(self, window_size):
    self.window = window_size

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index

    
    x = self.x_index_address[index]
    y = self.y_index_address[index]


    return x , y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples -1  


class Train_TEPS_split(TEPS_split):

  def __init__(self, window_size = 500, augmentations = False, max_samples = 8000000): 
    super().__init__(window_size, True, augmentations,max_samples)

class Test_TEPS_split(TEPS_split):

  def __init__(self, window_size = 500, augmentations = False, max_samples = 4800000): 
    super().__init__(window_size,  False , augmentations,max_samples)


class TEPS_split_binary(Dataset):
  def __init__(self, window_size, files : list , augmentations, max_samples):


    self.features = 52
    self.n_classes = 2
    self.samples_per_class = [0]* self.n_classes
    self.max_samples = max_samples

    path = "/home/snaags/scripts/datasets/TEPS/split/"
    data = []
    self.x_index_address = {}
    self.y_index_address = {}
    self.window = window_size
    while len(files) and min(self.samples_per_class) < max_samples:
      data.append(np.reshape(np.load(path+files.pop(random.randint(0, len(files)-1))),(-1,53)))
    self.current_index = 0
    use_all = True
  
    for batch in data:
      batch_x = batch[:,1:]
      batch_y = batch[:,0]
      self.add_to_dataset(batch_x,batch_y)

    self.n_samples = self.current_index #* len(self.augmentations)
 
  def add_to_dataset(self,x,y):
    y = np.where(y > 0, 1, 0) #Binarize Classes
    if self.samples_per_class[1]  < self.max_samples or np.unique(y)[0] == 0:
      self.x_index_address[self.current_index] = torch.from_numpy(x)
      self.y_index_address[self.current_index] = torch.from_numpy(y)
   
      if y[0] == 0:
        self.samples_per_class[0] += x.shape[0]
      else:
        self.samples_per_class[1] += x.shape[0]
      self.current_index += x.shape[0] - self.window

  def get_n_samples_per_class(self):
    for i in range(self.n_classes):
      print("Samples in class {}: {}".format(i, self.samples_per_class[i]))    

  def set_window_size(self, window_size):
    self.window = window_size

  def __getitem__(self, index):
    #index_address keys are the first usable index in a batch for the window size
    #this means the index

    for i in self.x_index_address:
      if (index - i) < 0:
        break
      else:
        addr = i

    index -= addr 
    x = self.x_index_address[addr]
    y = self.y_index_address[addr]

    x = x[index:index+self.window]
    y = y[index+self.window-1]

    return x.reshape( 52,self.window ) , y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples

  def __len__(self):
    return self.n_samples - self.n_samples%self.window


class Train_TEPS_split_binary(TEPS_split_binary):

  def __init__(self, train_files, window_size = 200, augmentations = True, max_samples = 250000): 
    super().__init__(window_size, train_files, augmentations,max_samples)

class Test_TEPS_split_binary(TEPS_split_binary):

  def __init__(self,test_files, window_size = 200, augmentations = False, max_samples = 480000): 
    super().__init__(window_size, test_files , augmentations,max_samples)


class Subset(Dataset):
    """
    Subset of a dataset at specified indices.
    Arguments:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """
    def __init__(self, dataset, indices):
        self.dataset = dataset
        self.indices = indices

    def __len__(self):
        if self.indices.shape == ():
            return 1
        else:
            return len(self.indices)
    def get_n_classes(self):
      return self.dataset.get_n_classes()
    def get_n_features(self):
      return self.dataset.get_n_features()
    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]



import numpy as np 
import time
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model, auto_train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute( ID = None, configs=None , gpus=None , res = None  , config = None):
  device = None
  print("Starting process: {}".format(ID))
  while not configs.empty():
    try:
      if device == None:
        device = gpus.get(timeout = 10)
      config = configs.get(timeout = 10)
    except Empty:
      if device != None:
        gpus.put(device)
      
    except:
      torch.cuda.empty_cache()
      if device != None:
        gpus.put(device)
      return
    if config != None:
      print("Got Configuration!")

    if device != None:
      print("Starting config with device: {}".format(device))
      complete = False
      crashes = 0
      acc, rec =  _compute(hyperparameter = config , cuda_device = device)
      while not complete:
        try:
          
          complete = True
        except:
          crashes +=1
          print("Model crash: {} ".format(crashes))
          time.sleep(60)
          if crashes == 2:
            print("Final Crash giving score of zero")
            results = {"accuracy": [0,0,0,0],"recall": [0,0,0,0]}
            complete = True
      res.put([config , acc, rec ]) 

  torch.cuda.empty_cache()
   
def _compute(hyperparameter,budget = 1, in_model = None , train_path = None,  test_path = None, cuda_device = None,plot_queue = None, model_id = None, binary = True):
  ### Configuration 
  THRESHOLD = 0.4 #Cut off for classification
  batch_size = 8
  result_dict = {"accuracy" : [],"recall" : []}
  if cuda_device == None:
     cuda_device = 1# torch.cuda.current_device()
  
  hpo = {'channels': 64, 'lr': 0.0025170869707739693, 'p': 0.0, 'epochs': 100, 'layers': 3}
  hpo.update(hyperparameter)
  ##Set up augmentations##
  jitter = aug.Jitter(device = cuda_device,sigma = 0.125, rate = 0.5)
  crop = aug.Crop(device = cuda_device, rate = 0.8, crop_min = 0.2 , crop_max = 0.98)
  scaling = aug.Scaling(device = cuda_device)
  window_warp = aug.WindowWarp(device = cuda_device,rate = 0.9)
  cut_out = aug.CutOut(device = cuda_device)
  mix_up = aug.MixUp(device = cuda_device,m = 0.3, rate = 0.2)
  augmentations = [jitter,crop,scaling, window_warp, cut_out]

  dataset_train = Train_TEPS(samples_per_class = 100,augmentations = augmentations,device = cuda_device,one_hot = False,PATH = "/home/cmackinnon/scripts/datasets/TEPS/split_HPO_samples_training/",binary = binary)
  #dataset_val = Train_TEPS(binary = binary,samples_per_class = 200,device = cuda_device,samples_per_epoch = 1,one_hot = False)
  #dataset_test = Test_TEPS(binary = binary,samples_per_class = 500,device = cuda_device,one_hot = False)
  #dataset_test_full = Test_TEPS()
  torch.cuda.set_device(cuda_device)

  print("Cuda Device Value: ", cuda_device)
  gen = config_space_2_DARTS(hyperparameter,reduction = True)

  
  n_classes = dataset_train.get_n_classes()
  inner = 5
  budget = 2
  multibatch = False
  for rep in range(budget):
    kfold = KFold(n_splits = inner, shuffle = True)
    for fold, (train_ids, test_ids) in enumerate(kfold.split(dataset_train,y = dataset_train.get_true_labels())):
      
      print('---Fold No.--{}-{}--------------------'.format(rep,fold))
      dataset_train.update_augmentation(augmentations)
      torch.cuda.empty_cache()
      # Sample elements randomly from a given list of ids, no replacement.
      train_subsampler = torch.utils.data.SubsetRandomSampler(train_ids)
      test_subsampler = torch.utils.data.SubsetRandomSampler(test_ids)
      
      # Define data loaders for training and testing data in this fold
      trainloader = torch.utils.data.DataLoader(
                        dataset_train, collate_fn = collate_fn_padd,
                        batch_size=batch_size, sampler=train_subsampler,drop_last = True)
      testloader = torch.utils.data.DataLoader(
                        dataset_train, collate_fn = collate_fn_padd,
                        batch_size=batch_size, sampler=test_subsampler,drop_last = True)
      val_evaluator = Evaluator(batch_size, n_classes,cuda_device,testloader = testloader) 
      evaluator = Evaluator(batch_size, n_classes,cuda_device,testloader = testloader) 

      model = NetworkMain(dataset_train.get_n_features(),hpo["channels"],num_classes= dataset_train.get_n_classes() , layers = hpo["layers"], auxiliary = False,drop_prob = hpo["p"], genotype = gen, binary = binary)
      model = model.cuda(device = cuda_device)
      """
      ### Train the model
      """
      _ = train_model(model , hpo, trainloader, hpo["epochs"], batch_size , cuda_device, graph = plot_queue, binary = binary,evaluator=val_evaluator,logger = False) 
      ##with aug
      #logger = train_model(model , hpo, trainloader_aug , hpo["epochs"], batch_size , cuda_device, graph = plot_queue, binary = binary,evaluator=val_evaluator,logger = logger) 
      """
      ### Test the model
      """
      dataset_train.disable_augmentation()
      evaluator.forward_pass(model, testloader,binary,n_iter = 1)
      #evaluator.predictions_threshold_matrix(binary)
      evaluator.predictions(model_is_binary = binary , THRESHOLD = THRESHOLD)
      evaluator.ROC()

      ### Get Metrics
      total = evaluator.T()
      acc  =  evaluator.T_ACC()
      recall = evaluator.TPR(1)
      recall_total = evaluator.P(1)
      #sup = evaluator.sup_loss(model, testloader)
      print("Accuracy: ", "%.4f" % ((acc)*100), "%")
      result_dict["accuracy"].append(acc)
      result_dict["recall"].append(recall)
      evaluator.reset_cm()
      ### Save Model
      def save_model(model, hpo):
        model_zoo = "{}/scripts/model_zoo/".format(os.environ["HOME"])
        torch.save(model.state_dict() , model_zoo+"-Acc-{}-Rec-{}".format(acc, recall))
        save_obj( hpo , model_zoo+"hps/"+"-Acc-{}-Rec-{}".format(acc , recall) )


  print("Final Scores -- ACC: {} -- REC: {}".format(acc, recall))
  return np.mean(result_dict["accuracy"]), np.mean(result_dict["recall"])

def report(l : list, name : str):
  std = np.std(l)
  avg = np.mean(l)
  print("{}: {} (+- {})".format(name,avg,std))

if __name__ == "__main__":
  hpo = {'channels': 32, 'lr': 0.0025170869707739693, 'p': 0.15, 'epochs': 100, 'layers': 3}
  hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 1, 'normal_index_1_1': 1, 'normal_index_2_0': 3, 'normal_index_2_1': 3, 'normal_index_3_0': 2, 'normal_index_3_1': 2, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'sep_conv_7x7', 'normal_node_1_0': 'sep_conv_5x5', 'normal_node_1_1': 'sep_conv_5x5', 'normal_node_2_0': 'sep_conv_7x7', 'normal_node_2_1': 'avg_pool_3x3', 'normal_node_3_0': 'skip_connect', 'normal_node_3_1': 'sep_conv_5x5', 'reduction_index_0_0': 1, 'reduction_index_0_1': 0, 'reduction_index_1_0': 0, 'reduction_index_1_1': 1, 'reduction_index_2_0': 1, 'reduction_index_2_1': 2, 'reduction_index_3_0': 4, 'reduction_index_3_1': 1, 'reduction_node_0_0': 'sep_conv_3x3', 'reduction_node_0_1': 'dil_conv_5x5', 'reduction_node_1_0': 'none', 'reduction_node_1_1': 'max_pool_3x3', 'reduction_node_2_0': 'dil_conv_3x3', 'reduction_node_2_1': 'avg_pool_3x3', 'reduction_node_3_0': 'none', 'reduction_node_3_1': 'sep_conv_5x5'}
  hyperparameter={
  'normal_index_0_0': 1,
  'normal_index_0_1': 0,
  'normal_index_1_0': 0,
  'normal_index_1_1': 2,
  'normal_index_2_0': 1,
  'normal_index_2_1': 2,
  'normal_index_3_0': 3,
  'normal_index_3_1': 0,
  'normal_node_0_0': 'max_pool_3x3',
  'normal_node_0_1': 'sep_conv_3x3',
  'normal_node_1_0': 'skip_connect',
  'normal_node_1_1': 'skip_connect',
  'normal_node_2_0': 'sep_conv_7x7',
  'normal_node_2_1': 'sep_conv_5x5',
  'normal_node_3_0': 'skip_connect',
  'normal_node_3_1': 'sep_conv_7x7',
  'reduction_index_0_0': 1,
  'reduction_index_0_1': 0,
  'reduction_index_1_0': 0,
  'reduction_index_1_1': 2,
  'reduction_index_2_0': 0,
  'reduction_index_2_1': 2,
  'reduction_index_3_0': 0,
  'reduction_index_3_1': 4,
  'reduction_node_0_0': 'none',
  'reduction_node_0_1': 'none',
  'reduction_node_1_0': 'none',
  'reduction_node_1_1': 'dil_conv_3x3',
  'reduction_node_2_0': 'avg_pool_3x3',
  'reduction_node_2_1': 'dil_conv_5x5',
  'reduction_node_3_0': 'max_pool_3x3',
  'reduction_node_3_1': 'sep_conv_3x3'}
  hyperparameter={'normal_index_0_0': 0,
  'normal_index_0_1': 1,
  'normal_index_1_0': 0,
  'normal_index_1_1': 1,
  'normal_index_2_0': 2,
  'normal_index_2_1': 0,
  'normal_index_3_0': 2,
  'normal_index_3_1': 0,
  'normal_node_0_0': 'dil_conv_3x3',
  'normal_node_0_1': 'avg_pool_3x3',
  'normal_node_1_0': 'skip_connect',
  'normal_node_1_1': 'none',
  'normal_node_2_0': 'sep_conv_5x5',
  'normal_node_2_1': 'avg_pool_3x3',
  'normal_node_3_0': 'dil_conv_3x3',
  'normal_node_3_1': 'dil_conv_5x5',
  'reduction_index_0_0': 1,
  'reduction_index_0_1': 0,
  'reduction_index_1_0': 0,
  'reduction_index_1_1': 0,
  'reduction_index_2_0': 0,
  'reduction_index_2_1': 0,
  'reduction_index_3_0': 1,
  'reduction_index_3_1': 4,
  'reduction_node_0_0': 'max_pool_3x3',
  'reduction_node_0_1': 'sep_conv_7x7',
  'reduction_node_1_0': 'skip_connect',
  'reduction_node_1_1': 'sep_conv_5x5',
  'reduction_node_2_0': 'avg_pool_3x3',
  'reduction_node_2_1': 'sep_conv_7x7',
  'reduction_node_3_0': 'max_pool_3x3',
  'reduction_node_3_1': 'none'}
  
  a_list = []
  r_list = [] 
  for i in range(1):
    a = _compute(hyperparameter, binary = True)
    report(a["accuracy"],"Accuracy")
    report(a["recall"],"Recall")
  exit()


  




import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
import os
import csv
import time 
from HPO.utils.ConfigStruct import Parameter, Cumulative_Integer_Struct, LTP_Parameter 

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""

def init_config():

  cs = CS.ConfigurationSpace()

  conv_ops= [ 
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'skip_connect',
    'sep_conv_3x3',
    'sep_conv_5x5',
    'sep_conv_7x7',
    'dil_conv_3x3',
    'dil_conv_5x5'
  ]
  

  ###DARTS###
  normal_node_0_0 = CSH.CategoricalHyperparameter('normal_node_0_0', choices=conv_ops)
  normal_node_0_1 = CSH.CategoricalHyperparameter('normal_node_0_1', choices=conv_ops)
  normal_index_0_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_0", lower = 0, upper = 1)
  normal_index_0_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_1", lower = 0, upper = 1)

  normal_node_1_0 = CSH.CategoricalHyperparameter('normal_node_1_0', choices=conv_ops)
  normal_node_1_1 = CSH.CategoricalHyperparameter('normal_node_1_1', choices=conv_ops)
  normal_index_1_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_0", lower = 0, upper = 2)
  normal_index_1_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_1", lower = 0, upper = 2)

  normal_node_2_0 = CSH.CategoricalHyperparameter('normal_node_2_0', choices=conv_ops)
  normal_node_2_1 = CSH.CategoricalHyperparameter('normal_node_2_1', choices=conv_ops)
  normal_index_2_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_0", lower = 0, upper = 3)
  normal_index_2_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_1", lower = 0, upper = 3)

  normal_node_3_0 = CSH.CategoricalHyperparameter('normal_node_3_0', choices=conv_ops)
  normal_node_3_1 = CSH.CategoricalHyperparameter('normal_node_3_1', choices=conv_ops)
  normal_index_3_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_0", lower = 0, upper = 4)
  normal_index_3_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_1", lower = 0, upper = 4)


  ###Optimiser###
  lr =CSH.UniformFloatHyperparameter(name = "lr",			lower = 0.000001  ,upper = 0.05)
  p =CSH.UniformFloatHyperparameter(name = "p",			lower = 0.01 ,upper = 0.3 )
  epochs = CSH.UniformIntegerHyperparameter(name = "epochs", lower = 4, upper = 150)
  layers = CSH.UniformIntegerHyperparameter(name = "layers", lower = 2, upper = 5)
  T_0 = CSH.UniformIntegerHyperparameter(name = "T_0", lower = 1, upper = 20)
  T_mult = CSH.UniformIntegerHyperparameter(name = "T_mult", lower = 1, upper = 3)
  channels = CSH.UniformIntegerHyperparameter(name = "channels", lower = 10, upper = 40)
  



  ##AugParameters 

  jitter = CSH.UniformFloatHyperparameter(name = "jitter",      lower = 0.001  ,upper = 0.5)
  scaling = CSH.UniformFloatHyperparameter(name = "scaling",      lower = 0.001  ,upper = 0.5)
  window_warp_num= CSH.UniformIntegerHyperparameter(name = "window_warp_num",     lower = 2  ,upper = 10)
  crop = CSH.UniformFloatHyperparameter(name = "crop",      lower = 0.05  ,upper = 0.9)
  mix_up = CSH.UniformFloatHyperparameter(name = "mix_up",      lower = 0.01  ,upper = 0.9)
  cut_out = CSH.UniformFloatHyperparameter(name = "cut_out",      lower = 0.05  ,upper = 0.9)
  cut_mix = CSH.UniformFloatHyperparameter(name = "cut_mix",      lower = 0.05  ,upper = 0.9)

  crop_rate = CSH.NormalFloatHyperparameter(name = "crop_rate", lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  jitter_rate= CSH.NormalFloatHyperparameter(name = "jitter_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  scaling_rate= CSH.NormalFloatHyperparameter(name = "scaling_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  window_warp_rate= CSH.NormalFloatHyperparameter(name = "window_warp_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  mix_up_rate= CSH.NormalFloatHyperparameter(name = "mix_up_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  cut_out_rate= CSH.NormalFloatHyperparameter(name = "cut_out_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
  cut_mix_rate= CSH.NormalFloatHyperparameter(name = "cut_mix_rate",lower = 0.0  ,mu = 0.5 , sigma = 0.5,upper = 5)
    ###Topology Definition]###



    ###Topology Definition]###
  
  hp_list = [
        lr,
        p,
        jitter ,
        scaling ,
        crop ,
        mix_up ,
        cut_out ,
        crop_rate ,
        jitter_rate,
        scaling_rate,
        window_warp_rate,
        mix_up_rate,
        cut_out_rate]

  cs.add_hyperparameters(hp_list)
  return cs

if __name__ == "__main__":
  from HPO.utils.DARTS_utils import config_space_2_DARTS
  configS = init_config()
  print(configS.get_hyperparameters())
  c = configS.sample_configuration()
  print(c)
  print(config_space_2_DARTS(c))



import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt

class SHAR(Dataset):

  def __init__(self,path, augmentation = None, cuda_device = 0,alt_format = False,**kwargs):
    self.cuda_device = cuda_device
    PATH_x = "{}{}".format(path,"data.npy")
    PATH_y = "{}{}".format(path,"labels.npy")
    PATH_group = "{}{}".format(path,"groups.npy")
    #data = np.swapaxes(data,1,2)
    
    self.x = torch.from_numpy(np.load(PATH_x))
    self.groups = np.load(PATH_group)
    self.x  = self.x.to(cuda_device).float()
    self.y = np.load("{}".format(PATH_y)) -1 
    self.augmentation = augmentation
    if kwargs["binary"]:
      self.y = np.where(self.y != 0, 1,0)
    self.y = torch.from_numpy(self.y).to(cuda_device).long()
  def get_labels(self):
    return self.y
  def disable_augmentation(self):
    self.augmentation = False
  def enable_augmentation(self,aug):
    self.augmentation = aug

  def get_groups(self, train_id, test_id):
      train_groups = self.groups[train_id]
      test_groups = self.groups[test_id]
      print("Groups in train are: {} with a total length: {}".format(np.unique(train_groups),len(train_groups)))
      print("Groups in test are: {} with a total length: {}".format(np.unique(test_groups),len(test_groups)))
  def __getitem__(self, index):
    x, y = self.x[index], self.y[index]
    if self.augmentation:
      for func in self.augmentation:
        x,y = func(x,y)
    return x,y
  
  def get_n_classes(self):
    print("n classes: {}".format(len(torch.unique(self.y))))
    return len(torch.unique(self.y))
  def get_n_samples(self):
    return self.x.shape[0]

  def get_proportions(self):
    return 0.2
  def get_n_features(self):
    return self.x.shape[1]
  def get_length(self):
    return self.x.shape[2]

  def __len__(self):
    return self.x.shape[0]


class SHAR_TRAIN(SHAR):
  def __init__(self,path = "/home/cmackinnon/scripts/datasets/SHAR/train/", augmentation = None, cuda_device = 0,**kwargs):
     super(SHAR_TRAIN,self).__init__(path =path,augmentation = augmentation, cuda_device = cuda_device, **kwargs)

class SHAR_TEST(SHAR):
  def __init__(self,path = "/home/cmackinnon/scripts/datasets/SHAR/test/", augmentation = None, cuda_device = 0,**kwargs):
     super(SHAR_TEST,self).__init__(path = path,augmentation = augmentation, cuda_device = cuda_device, **kwargs)

class SHAR_PARTITION(SHAR):
  def __init__(self,path = "/home/cmackinnon/scripts/datasets/SHAR/test/", augmentation = None, cuda_device = 0,**kwargs):
     super(SHAR_TEST,self).__init__(path = path,augmentation = augmentation, cuda_device = cuda_device, **kwargs)

class Full_SHAR(SHAR):
  def __init__(self, augmentation = None, cuda_device = 0,**kwargs):
     super(Full_SHAR,self).__init__(path = "/home/cmackinnon/scripts/datasets/SHAR/",augmentation = augmentation, cuda_device = cuda_device, **kwargs)



"""
class SHAR_TEST(Dataset):

  def __init__(self, augmentation = None, cuda_device = 0,**kwargs):
    PATH = "/home/cmackinnon/scripts/datasets/SHAR/"
    self.cuda_device = cuda_device
    PATH_x = "/home/cmackinnon/scripts/datasets/SHAR/test/data.npy"#UCI HAR Dataset/train/Inertial Signals"
    PATH_y = "/home/cmackinnon/scripts/datasets/SHAR/test/labels.npy"
    PATH_group = "/home/cmackinnon/scripts/datasets/SHAR/test/groups.npy"
    #data = np.swapaxes(data,1,2)
    
    self.x = torch.from_numpy(np.load(PATH_x))
    self.groups = np.load(PATH_group)
    self.x  = self.x.cuda(cuda_device).float()
    self.y = np.load("{}".format(PATH_y)) -1 
    self.augmentation = augmentation
    if kwargs["binary"]:
      self.y = np.where(self.y != 0, 1,0)
    self.y = torch.from_numpy(self.y).cuda(cuda_device).long()
  def get_labels(self):
    return self.y
  def disable_augmentation(self):
    self.augmentation = False
  def enable_augmentation(self,aug):
    self.augmentation = aug

  def get_groups(self, train_id, test_id):
      train_groups = self.groups[train_id]
      test_groups = self.groups[test_id]
      print("Groups in train are: {} with a total length: {}".format(np.unique(train_groups),len(train_groups)))
      print("Groups in test are: {} with a total length: {}".format(np.unique(test_groups),len(test_groups)))
  def __getitem__(self, index):
    x, y = self.x[index], self.y[index]
    if self.augmentation:
      for func in self.augmentation:
        x,y = func(x,y)
    return x,y
  
  def get_n_classes(self):
    print("n classes: {}".format(len(torch.unique(self.y))))
    return len(torch.unique(self.y))
  def get_n_samples(self):
    return self.x.shape[0]
  def get_n_features(self):
    return self.x.shape[1]
  def get_length(self):
    return self.x.shape[2]

  def __len__(self):
    return self.x.shape[0]
"""



from HPO.utils.graph_utils import gen_iter
from HPO.searchspaces.utils import *
import copy
import ConfigSpace.hyperparameters as CSH
import ConfigSpace as CS
import time
import json
import networkx as nx


def is_fully_connected(G: nx.DiGraph) -> bool:
    # Count nodes with in-degree of 0 and out-degree of 0
    no_input_nodes = sum(1 for _, indeg in G.in_degree() if indeg == 0)
    no_output_nodes = sum(1 for _, outdeg in G.out_degree() if outdeg == 0)

    # Check if there is exactly one node with no inputs and one node with no outputs
    return (no_input_nodes == 1) and (no_output_nodes == 1)

def is_valid_network(G, input_node, output_node):
    # Check if there's only one input and one output node
    input_count = 0
    output_count = 0

    if not "S" in G.nodes or not "T" in G.nodes:
      return False
    """
    for node, degree in G.degree():
        if degree == 1:
            if G.out_degree(node) == 1:
                input_count += 1
                if input_count > 1:
                    return False
            elif G.in_degree(node) == 1:
                output_count += 1
                if output_count > 1:
                    return False
    
    if input_count != 1 or output_count != 1:
        return False
    """
    # Check if the graph is acyclic
    if not nx.is_directed_acyclic_graph(G):
        return False

    # Check if all nodes are part of a path between input and output nodes
    for n in G.nodes:
      try:
          # If there's a path from the input to output, then all nodes are connected
          _ = nx.shortest_path(G, n, output_node)
          _ = nx.shortest_path(G, input_node,n)
      except nx.NetworkXNoPath:
          return False

    return True

def get_ops(n_ops = 30):

  cs = CS.ConfigurationSpace()

  conv_ops= [ 
    'max_pool',
    'avg_pool',
    'skip_connect',
    'point_conv' ,
    'depth_conv',
    'gelu',
    'batch_norm']
    
    #'max_pool_31x31',
    #'avg_pool_31x31',
    #'max_pool_64x64',
    #'avg_pool_64x64',
    #'depth_conv_15',
    #'depth_conv_29' ,
    #'depth_conv_61' ,
    #'depth_conv_101',
    #'depth_conv_201',
    #'SE_8',
    #'SE_16',
    #'attention_space',
    #'attention_channel']
  


  ###DARTS###
  hp_list = []
  for i in range(n_ops):  
    hp_list.append(CSH.CategoricalHyperparameter('op_{}'.format(i), choices=conv_ops))
    hp_list.append(CSH.UniformIntegerHyperparameter('op_{}_kernel'.format(i), lower = 2 , upper = 4))#kernel
    hp_list.append(CSH.UniformIntegerHyperparameter('op_{}_stride'.format(i), lower = 1 , upper = 4))#stride
    hp_list.append(CSH.UniformIntegerHyperparameter('op_{}_dil'.format(i), lower = 0 , upper = 4))#dilation
    #hp_list.append(CSH.UniformIntegerHyperparameter('op_{}_channels'.format(i), lower = 2 , upper = 5))#channels
  cs.add_hyperparameters(hp_list)
  return cs





class GraphConfigSpace:
  def __init__(self,JSON):
    with open(JSON) as conf:
      self.data = json.load(conf)["ARCHITECTURE_CONFIG"]
    #self.data = JSON["ARCHITECTURE_CONFIG"]
    self.g = nx.DiGraph
    self.n_operations = 32
    self.edge_options = self.data["N_EDGES"]
    self.init_state = [("S",1),(1,"T")]
  def sample_configuration(self,n_samples=1):
    samples = []
    while len(samples) < n_samples:
      graph = copy.copy(self.init_state)
      self.n_operations = random.choice(self.edge_options)
      while len(graph) < self.n_operations:
        rate = 0.5 
        self.g = nx.DiGraph()
        self.g.add_edges_from(graph)
        graph = gen_iter(graph,self.g,rate)
      self.g.add_edges_from(graph)
      model_stride = random.choice(self.data["STRIDE_RATE"])
      model_stride_channel_ratio = random.choice(self.data["CHANNEL_DEPTH_RATE"])
      if self.data["STRIDE_VARIABLE"]:
        stride = random.randint(1,self.data["STRIDE_COUNT"])
      else:
        stride = self.data["STRIDE_COUNT"]
      ops = generate_op_names(self.g)
      ops = random_ops_unweighted(ops, self.data)
      ops = random_activation_unweighted(ops,self.data)
      ops = random_normalisation_unweighted(ops,self.data)
      ops = random_combine_unweighted(ops,self.data)
      
      del ops["T_stride"]
      del ops["T_channel_ratio"]
      del ops["S_stride"]
      del ops["S_channel_ratio"]
      ops = random_strides(ops,stride,self.data,2,model_stride_channel_ratio)
      #ops["s_rate"] = model_stride
      #ops["s_c_ratio"] = model_stride_channel_ratio
      #ops["stem"] = random.choice(self.data["STEM_SIZE"])
      samples.append(Graph(graph, self.g.nodes,copy.copy(ops),self.data))
    return samples
  
  def mutate_graph(self,m,steps = 1):
    edges = m["graph"]
    ops = m["ops"]
    g = nx.DiGraph()
    g.add_edges_from(edges )
    if "ID" in m:
      model = Graph(edges,g.nodes,ops,self.data,parent = m["ID"])
    elif "parent" in m["ops"]:
      model = Graph(edges,g.nodes,ops,self.data,parent = m["ops"]["parent"])
    else:
      model = Graph(edges,g.nodes,ops,self.data)
    result = False
    while result == False:
      result = self.check_valid(copy.deepcopy(model))
    #print("Time to find valid model: ",time.time()- s)
    if steps>0:
      return self.mutate_graph(result(),steps-1)
    return result

  def check_valid(self,model):
    #MAKE RANDOM EDIT AND GENERATE NX GRAPH
    model.get_random_edit()()
    graph = model()["graph"]
    g = nx.DiGraph()
    g.add_edges_from(graph)

    #CHECK GRAPH IS VALID
    valid = nx.is_directed_acyclic_graph(g) 

    if is_valid_network(g, "S", "T"):
      return model
    else:
      return False


from functools import wraps

def log_graph_change(func):
    @wraps(func)
    def wrapper(G, *args, **kwargs):
        edges_before = set(G.edges)
        nodes_before = set(G.nodes)
        result = func(G, *args, **kwargs)
        edges_after = set(G.edges)
        nodes_after = set(G.nodes)

        added_edges = edges_after - edges_before
        removed_edges = edges_before - edges_after
        if added_edges:
            print(f"{func.__name__} added: {added_edges}")
        if removed_edges:
            print(f"{func.__name__} removed: {removed_edges}")

        added_nodes = nodes_after - nodes_before
        removed_nodes = nodes_before - nodes_after

        if added_nodes:
            print(f"{func.__name__} added: {added_nodes}")
        if removed_nodes:
            print(f"{func.__name__} removed: {removed_nodes}")

        return result

    return wrapper




class Edge:
  def __init__(self,source,end,operation = None):
    self.source = source
    self.end = end
    self.operation = operation

  def __call__(self):
    out_dict = {}
    out_dict["{}_{}_OP".format(self.source,self.end)] = self.operation
    return out_dict

  def get_edge(self):
    return (self.source,self.end)

  def update(self,value):
    if type(self.source) == int:
      self.source += value
    if type(self.end) == int:   
      self.end += value

class Vertex:
  def __init__(self,ID,attributes : dict):
    for key, value in attributes.items():
        setattr(self, key, value)
    self.id = ID
    self.attributes = attributes

  def __call__(self):
    out_dict = {}
    for key, value in self.attributes.items():
      out_dict["{}_{}".format(self.id,key)] = value
    return out_dict

  def update(self,value):
    if type(self.id) == int:
      self.id += value


class Graph:
    def __init__(self,edges,nodes,ops,data,parent = None):
        self.vertices = {}
        self.data = data
        self.edges = {}
        self.parent = parent
        self.max_vertex = 1
        for e in edges:
          self.edges[e] = Edge(e[0],e[1],ops["{}_{}_OP".format(e[0],e[1])])
        for v in nodes:
          attr = {x.split("_")[-1] : ops[x] for x in ops if str(v) == x.split("_")[0] and not "OP" in x}
          self.vertices[v] = Vertex(v, attr)
          if type(v) != str:
            if v > self.max_vertex:
              self.max_vertex = v

    def __call__(self):
      ops = {k: v for d in self.vertices for k, v in self.vertices[d]().items()}
      ops.update({k: v for d in self.edges for k, v in self.edges[d]().items()})
      if self.parent != None:
        ops["parent"] = self.parent
      graph = [ self.edges[e].get_edge() for e in self.edges]
      return { "graph":graph,"ops":ops}


    def rename_vertices(self,value):
      for e in self.edges:
        self.edges[e].update(value)
      for v in self.vertices:
        self.vertices[v].update(value)


    def to_networkx(self):
        # Create a new directed graph using networkx
        G = nx.DiGraph()
        # Add nodes to the graph with their attributes
        for vertex_id, vertex in self.vertices.items():
            G.add_node(vertex_id, **vertex.attributes)
        # Add edges to the graph with their attributes
        for edge_key, edge in self.edges.items():
            G.add_edge(edge.source, edge.end, operation=edge.operation)
        return G
    def new_vertex(self):
      self.max_vertex += 1
      return self.max_vertex

    def get_total_stride(self):
      count = 1
      for v in self.vertices:
        if v != "S" or v != "T":
          if "stride" in self.vertices[v].attributes:
              count *= self.vertices[v].attributes["stride"]
      return count
    def get_random_edit(self):
      operations = [
        self.edge_split,
        self.edge_contraction,
        self.vertex_substitution,
        self.edge_insertion,
        self.edge_deletion,
        self.edge_substitution]
      return random.choice(operations)

    def get_vertex(self):
      return random.choice(list(self.vertices.keys()))

    def get_edge(self):
      e = random.choice(list(self.edges.keys()))
      return e[0], e[1]



    def get_dictionary(self):
      return self.__call__()


    def edge_split(self):
        u,v = self.get_edge()
        edge = self.edges.pop((u, v))
        new_vertex = self.new_vertex()
        self.vertices[new_vertex] = Vertex(new_vertex, self.generate_random_vertex())
        self.edges[(u, new_vertex)] = Edge(u, new_vertex, edge.operation)
        self.edges[(new_vertex, v)] = Edge(new_vertex, v, edge.operation)

    def edge_contraction(self):
        u,v = self.get_edge()
        edge_u_v = self.edges.pop((u, v))
        for (w, x) in list(self.edges.keys()):
            if x == v:
                edge_v_w = self.edges.pop((w, x))
                self.edges[(w, u)] = Edge(w, u, edge_v_w.operation)
            elif w == v:
                edge_v_w = self.edges.pop((w, x))
                self.edges[(u, x)] = Edge(u, x, edge_v_w.operation)
        #print(self.vertices.keys())
        #print(self.edges.keys())
        del self.vertices[v]

    def vertex_substitution(self):
        v = self.get_vertex()
        while v == "S" or v == "T":
          v = self.get_vertex()
        vertex = self.vertices[v]
        vertex.attributes = self.generate_random_vertex()

    def edge_insertion(self):
        u,v = self.get_vertex(), self.get_vertex()
        while (u, v) in self.edges:
          u,v = self.get_vertex(), self.get_vertex()
        self.edges[(u, v)] = Edge(u, v,  self.random_operation())

    def edge_deletion(self):
        u,v = self.get_edge()
        del self.edges[(u, v)]

    def edge_substitution(self):
        u,v = self.get_edge()
        edge = self.edges[(u, v)]
        edge.operation = self.random_operation()

    def random_operation(self):
      return random.choice(self.data["EDGE_OPERATIONS"])
    def generate_random_vertex(self):
      attr = {}
      attr["activation"]= random.choice(self.data["ACTIVATION_FUNCTIONS"])
      attr["combine"] = random.choice(self.data["COMBINE_OPERATIONS"])
      attr["normalisation"] = random.choice(self.data["NORMALISATION_FUNCTIONS"])
      #attr["channel_ratio"] = random.choice(self.data["CHANNEL_RATIOS"])
      attr["stride"] = random.choice(self.data["STRIDES"])
      if self.get_total_stride() >  self.data["STRIDE_COUNT"]:
        attr["stride"] = 1
      return attr




import csv

def load(FILENAME):
    scores = []
    recall = []
    config = []
    params = []
    with open( "{}".format(FILENAME) , newline = "") as csvfile:
        reader = csv.reader(csvfile, delimiter = ",")
        for row in reader:
            scores.append(float(row[0]))
            recall.append(float(row[1]))
            config.append(eval("".join(row[2])))
            if len(row) == 4:
               params.append(int(row[3])) 
    error = [1-x for x in scores]
    e_min = 1
    best_list = []
    for i in error:
      if i < e_min:
        e_min = i
      best_list.append(e_min)
    print("Best Score: {}".format(max(scores)))
    return {"scores":scores,"recall":recall,"config":config,"error":error,"best":best_list ,"params": params}



def plot_graph(edges,ops):
    g= nx.DiGraph()
    g.add_edges_from(edges)
    plt.figure(figsize = (50,10))
    for i, layer in enumerate(nx.topological_generations(g)):
        for n in layer:
            g.nodes[n]["layer"] = i
    pos = nx.multipartite_layout(g,subset_key="layer", align="vertical")
    for i in pos:
        temp= pos[i]
        temp[1] += random.uniform(-0.3,0.3)
        
    edge_labels = {e: ops[ "{}_{}_OP".format(e[0], e[1])] for e in g.edges()}
    res_dict =  propagate_resolution(edges, ops)
    nx.draw(
         g, edge_color='black',pos = pos , width=1, linewidths=5,
         node_size=2000, node_color='pink', alpha=0.9,font_size = 20,with_labels=True, labels={node: "{}({})".format(node,res_dict[node]) for node in g.nodes()}
         )
    
    nx.draw_networkx_edge_labels(
    g, pos,
    edge_labels=edge_labels,
    font_color='red',
    font_size = 15
)
    plt.axis('off')
    plt.savefig("test") 
    return g


import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
import os
import csv
import time 
from HPO.utils.ConfigStruct import Parameter, Cumulative_Integer_Struct, LTP_Parameter 

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""
 
def init_config():

  cs = CS.ConfigurationSpace()

  conv_ops= [ 
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'skip_connect',
    'sep_conv_3x3',
    'sep_conv_5x5',
    'sep_conv_7x7',
    'dil_conv_3x3',
    'dil_conv_5x5'
  ]
  

  ###DARTS###
  normal_node_0_0 = CSH.CategoricalHyperparameter('normal_node_0_0', choices=conv_ops)
  normal_node_0_1 = CSH.CategoricalHyperparameter('normal_node_0_1', choices=conv_ops)
  normal_index_0_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_0", lower = 0, upper = 1)
  normal_index_0_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_0_1", lower = 0, upper = 1)

  normal_node_1_0 = CSH.CategoricalHyperparameter('normal_node_1_0', choices=conv_ops)
  normal_node_1_1 = CSH.CategoricalHyperparameter('normal_node_1_1', choices=conv_ops)
  normal_index_1_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_0", lower = 0, upper = 2)
  normal_index_1_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_1_1", lower = 0, upper = 2)

  normal_node_2_0 = CSH.CategoricalHyperparameter('normal_node_2_0', choices=conv_ops)
  normal_node_2_1 = CSH.CategoricalHyperparameter('normal_node_2_1', choices=conv_ops)
  normal_index_2_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_0", lower = 0, upper = 3)
  normal_index_2_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_2_1", lower = 0, upper = 3)

  normal_node_3_0 = CSH.CategoricalHyperparameter('normal_node_3_0', choices=conv_ops)
  normal_node_3_1 = CSH.CategoricalHyperparameter('normal_node_3_1', choices=conv_ops)
  normal_index_3_0 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_0", lower = 0, upper = 4)
  normal_index_3_1 = CSH.UniformIntegerHyperparameter(name = "normal_index_3_1", lower = 0, upper = 4)


  reduction_node_0_0 = CSH.CategoricalHyperparameter('reduction_node_0_0', choices=conv_ops)
  reduction_node_0_1 = CSH.CategoricalHyperparameter('reduction_node_0_1', choices=conv_ops)
  reduction_index_0_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_0_0", lower = 0, upper = 1)
  reduction_index_0_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_0_1", lower = 0, upper = 1)

  reduction_node_1_0 = CSH.CategoricalHyperparameter('reduction_node_1_0', choices=conv_ops)
  reduction_node_1_1 = CSH.CategoricalHyperparameter('reduction_node_1_1', choices=conv_ops)
  reduction_index_1_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_1_0", lower = 0, upper = 2)
  reduction_index_1_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_1_1", lower = 0, upper = 2)

  reduction_node_2_0 = CSH.CategoricalHyperparameter('reduction_node_2_0', choices=conv_ops)
  reduction_node_2_1 = CSH.CategoricalHyperparameter('reduction_node_2_1', choices=conv_ops)
  reduction_index_2_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_2_0", lower = 0, upper = 3)
  reduction_index_2_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_2_1", lower = 0, upper = 3)

  reduction_node_3_0 = CSH.CategoricalHyperparameter('reduction_node_3_0', choices=conv_ops)
  reduction_node_3_1 = CSH.CategoricalHyperparameter('reduction_node_3_1', choices=conv_ops)
  reduction_index_3_0 = CSH.UniformIntegerHyperparameter(name = "reduction_index_3_0", lower = 0, upper = 4)
  reduction_index_3_1 = CSH.UniformIntegerHyperparameter(name = "reduction_index_3_1", lower = 0, upper = 4)


  ###Optimiser###



  lr =CSH.UniformFloatHyperparameter(name = "lr",			lower = 0.000001  ,upper = 0.05)
  p =CSH.UniformFloatHyperparameter(name = "p",			lower = 0.01 ,upper = 0.3 )
  epochs = CSH.UniformIntegerHyperparameter(name = "epochs", lower = 1, upper = 5)
  layers = CSH.UniformIntegerHyperparameter(name = "layers", lower = 3, upper = 6)
  c1 = CSH.UniformFloatHyperparameter(name = "c1_weight" , lower = 1,upper = 3)
  batch_size = CSH.UniformIntegerHyperparameter(name = "batch_size", lower = 2, upper = 16)
  T_0 = CSH.UniformIntegerHyperparameter(name = "T_0", lower = 2, upper = 10)
  T_mult = CSH.UniformIntegerHyperparameter(name = "T_mult", lower = 1, upper = 3)
  channels = CSH.UniformIntegerHyperparameter(name = "channels", lower = 16, upper = 32)
  augmentations = CSH.UniformIntegerHyperparameter(name = "augmentations", lower = 0, upper = 20)
  
    ###Topology Definition]###
  
  hp_list = [
        layers,
        channels,
        normal_node_0_0 ,
        normal_node_0_1 ,
        normal_index_0_0,
        normal_index_0_1,
        normal_node_1_0 ,
        normal_node_1_1 ,
        normal_index_1_0,
        normal_index_1_1,
        normal_node_2_0 ,
        normal_node_2_1 ,
        normal_index_2_0,
        normal_index_2_1,
        normal_node_3_0 ,
        normal_node_3_1 ,
        normal_index_3_0, 
        normal_index_3_1,
        reduction_node_0_0 ,
        reduction_node_0_1 ,
        reduction_index_0_0,
        reduction_index_0_1,
        reduction_node_1_0 ,
        reduction_node_1_1 ,
        reduction_index_1_0,
        reduction_index_1_1,
        reduction_node_2_0 ,
        reduction_node_2_1 ,
        reduction_index_2_0,
        reduction_index_2_1,
        reduction_node_3_0 ,
        reduction_node_3_1 ,
        reduction_index_3_0, 
        reduction_index_3_1]
  cs.add_hyperparameters(hp_list)
  return cs

if __name__ == "__main__":
  from HPO.utils.DARTS_utils import config_space_2_DARTS
  configS = init_config()	
  print(configS.get_hyperparameters())
  c = configS.sample_configuration()
  print(c)
  print(config_space_2_DARTS(c))



import numpy as np
import torch.nn.functional as F
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt

class HAR(Dataset):

  def __init__(self, train: bool, augmentation = None, cuda_device = 0,**kwargs):
    PATH = "/home/cmackinnon/scripts/datasets/HAR/UCI HAR Dataset/"
    self.cuda_device = cuda_device
    if train:
      PATH_x = "/home/cmackinnon/scripts/datasets/HAR/train_x.npy"#UCI HAR Dataset/train/Inertial Signals"
      PATH_y = "/home/cmackinnon/scripts/datasets/HAR/UCI HAR Dataset/train/y_train.txt"
    else:
      PATH_x = "/home/cmackinnon/scripts/datasets/HAR/test_x.npy"#UCI HAR Dataset/test/Inertial Signals"
      PATH_y = "/home/cmackinnon/scripts/datasets/HAR/UCI HAR Dataset/test/y_test.txt"
    #data = np.swapaxes(data,1,2)
    
    self.x = torch.from_numpy(np.load(PATH_x))

    self.x  = self.x.cuda(cuda_device).float()
    self.y = np.loadtxt("{}".format(PATH_y)) -1
    self.augmentation = augmentation
    if kwargs["binary"]:
      self.y = np.where(self.y != 0, 1,0)
    self.y = torch.from_numpy(self.y).cuda(cuda_device).long()
  def get_labels(self):
    return self.y
  def disable_augmentation(self):
    self.augmentation = False
  def enable_augmentation(self,aug):
    self.augmentation = aug

  def __getitem__(self, index):
    x, y = self.x[index], self.y[index]
    if self.augmentation:
      for func in self.augmentation:
        x,y = func(x,y)
    return x,y
  
  def get_n_classes(self):
    print("n classes: {}".format(len(torch.unique(self.y))))
    return len(torch.unique(self.y))
  def get_n_samples(self):
    return self.x.shape[0]
  def get_n_features(self):
    return self.x.shape[1]

  def __len__(self):
    return self.x.shape[0]


class Train_HAR(HAR):

  def __init__(self, augmentation = False,samples_per_class = None,binary = False,one_hot = True,samples_per_epoch = 1,PATH= None,device = None,sub_set_classes = None): 
    super(Train_HAR,self).__init__(True, augmentation,cuda_device = device,samples_per_class = samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,PATH = PATH,sub_set_classes = sub_set_classes)

class Test_HAR(HAR):

  def __init__(self, augmentation = False,samples_per_class = None,binary = False,one_hot = False,samples_per_epoch = 1,device = None,sub_set_classes = None): 
    super(Test_HAR,self).__init__(False ,cuda_device = device, augmentation = augmentation,samples_per_class = samples_per_class,binary = binary ,one_hot = one_hot,samples_per_epoch = samples_per_epoch,sub_set_classes = sub_set_classes)



import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
import os
import csv
import time 
from HPO.utils.ConfigStruct import Parameter, Cumulative_Integer_Struct, LTP_Parameter 

"""	TODO
Seperate Pooling and Convolution Layers
Add more convolution operations (kernalSize and maybe stride)
"""

def init_config():

  cs = CS.ConfigurationSpace()



  ###Optimiser###


 
  lr =CSH.UniformFloatHyperparameter(name = "lr",			lower = 0.000001  ,upper = 0.1)
  p =CSH.UniformFloatHyperparameter(name = "p",			lower = 0.01 ,upper = 0.3 )
  epochs = CSH.UniformIntegerHyperparameter(name = "epochs", lower = 30, upper = 300)
  c1 = CSH.UniformFloatHyperparameter(name = "c1_weight" , lower = 1,upper = 5)
  T_0 = CSH.UniformIntegerHyperparameter(name = "T_0", lower = 1, upper = 10)
  T_mult = CSH.UniformIntegerHyperparameter(name = "T_mult", lower = 1, upper = 3)

  ##AugParameters 

  jitter = CSH.UniformFloatHyperparameter(name = "jitter",			lower = 0.001  ,upper = 0.5)
  scaling = CSH.UniformFloatHyperparameter(name = "scaling",			lower = 0.001  ,upper = 0.5)
  window_warp_num= CSH.UniformIntegerHyperparameter(name = "window_warp_num",			lower = 2  ,upper = 10)
  jitter_rate= CSH.UniformFloatHyperparameter(name = "jitter_rate",			lower = 0.05  ,upper = 0.9)
  scaling_rate= CSH.UniformFloatHyperparameter(name = "scaling_rate",			lower = 0.05  ,upper = 0.9)
  window_warp_rate= CSH.UniformFloatHyperparameter(name = "window_warp_rate",			lower = 0.05  ,upper = 0.9)
  
    ###Topology Definition]###
  
  hp_list = [
        c1,
        epochs,
        lr,
        p,
        T_0,
        T_mult,
        jitter,
        scaling,
        jitter_rate,
        scaling_rate,
        window_warp_rate,
        window_warp_num]
  cs.add_hyperparameters(hp_list)
  return cs

if __name__ == "__main__":
  from HPO.utils.DARTS_utils import config_space_2_DARTS
  configS = init_config()	
  print(configS.get_hyperparameters())
  c = configS.sample_configuration()
  print(c)
  print(config_space_2_DARTS(c))



import copy
from collections import namedtuple
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')



def config_space_2_DARTS(hyperparameters,reduction = False, N = 16):
  normal = [0] * N
  reduce = [0] * N
  for i in hyperparameters:
    if "normal" in i:
      idx = (int(i[-3])*2) + int(i[-1]) 
      if "index" in i:
        if type(normal[idx]) == int:
          normal[idx] = [0,hyperparameters[i]]
        else:
          normal[idx][1] = hyperparameters[i]
        
      if "node" in i:
        if type(normal[idx]) == int:
          normal[idx] = [hyperparameters[i], 0]
        else:
          normal[idx][0] = hyperparameters[i]
           
    if "reduction" in i:
      idx = (int(i[-3])*2) + int(i[-1]) 
      idx = (int(i[-3])*2) + int(i[-1]) 
      if "index" in i:
        if type(reduce[idx]) == int:
          reduce[idx] = [0,hyperparameters[i]]
        else:
          reduce[idx][1] = hyperparameters[i]
        
      if "node" in i:
        if type(reduce[idx]) == int:
          reduce[idx] = [hyperparameters[i], 0]
        else:
          reduce[idx][0] = hyperparameters[i]
  normal_concat = [x for x in range(2,len(normal)//2+2)]
  reduce_concat = [x for x in range(2,len(normal)//2+2)]
  if reduction == False:
    print("Reduction cell architecture disabled!")
    reduce = copy.copy(normal)
  for n,r in zip(normal, reduce):
    if n[1] in normal_concat:
      normal_concat.pop(normal_concat.index(n[1]))
    if r[1] in reduce_concat:
      reduce_concat.pop(reduce_concat.index(r[1]))
  normal = [tuple(x) for x in normal ]
  reduce = [tuple(x) for x in reduce ]
  return Genotype(normal , normal_concat, reduce, reduce_concat)



import torch
import torch.nn as nn
from HPO.utils.operations import *
#from HPO.utils.operations2d import *
from torch.autograd import Variable
from HPO.utils.utils import drop_path


class CellAttn(nn.Module):
  """
  genotype : holds cell data
  C_prev_prev : channels of k-2
  C_prev : channels of k-1
  C : channels of k (i.e. current layer)
  reduction : boolean whether this layer is a reduction cell
  reduction_prev : boolean of whether the last cell was a reduction cell
  """
  def __init__(self, genotype, C_prev_prev, C_prev, C, reduction, reduction_prev):
    super(CellAttn, self).__init__()

    if reduction_prev:
      self.preprocess0 = FactorizedReduce(C_prev_prev, C)
    else:
      self.preprocess0 = ReLUConvBN(C_prev_prev, C, 1, 1, 0)
    self.attn = Attention()
    self.preprocess1 = ReLUConvBN(C_prev, C, 1, 1, 0)
    self._channels = [C,C_prev,C_prev_prev] 
    self._reduc = [reduction,reduction_prev] 
    if reduction:
      op_names, indices = zip(*genotype.reduce)
      concat = genotype.reduce_concat
    else:
      op_names, indices = zip(*genotype.normal) #get operation names and indexs from genotype
      concat = genotype.normal_concat
    self._compile(C, op_names, indices, concat, reduction)

  def _compile(self, C, op_names, indices, concat, reduction):
    assert len(op_names) == len(indices)
    self._steps = len(op_names) // 2
    self._concat = concat
    self.multiplier = len(concat)

    self._ops = nn.ModuleList()
    for name, index in zip(op_names, indices):
      stride = 2 if reduction and index < 2 else 1
      op = OPS[name](C, stride, True)
      self._ops += [op]
    self._indices = indices

  def forward(self, s0, s1, drop_prob):
    s0 = self.preprocess0(s0)
    s1 = self.preprocess1(s1)
    
    states = [s0, s1]
    for i in range(self._steps):
      h1 = states[self._indices[2*i]]
      h2 = states[self._indices[2*i+1]]
      op1 = self._ops[2*i]
      op2 = self._ops[2*i+1]
    
      h1 = op1(h1)
      h2 = op2(h2)
      if self.training and drop_prob > 0.:
        if not isinstance(op1, Identity):
          h1 = drop_path(h1, drop_prob)
        if not isinstance(op2, Identity):
          h2 = drop_path(h2, drop_prob)
      if h1.shape[1] != h2.shape[1]:
        print(op1,op2)
        print("Channels k, k-1, k-2: {} {} {}".format(*self._channels))
        print("Reduction: {} - Previous Reduction: {} ".format(*self._reduc))
      
      s = self.attn(torch.cat((h1,h2),dim = 2))
      #print("Attenion Map Shape: {}".format(s.shape))
      states += [s]
    return torch.cat([states[i] for i in self._concat], dim=1)

class Cell(nn.Module):
  """
  genotype : holds cell data
  C_prev_prev : channels of k-2
  C_prev : channels of k-1
  C : channels of k (i.e. current layer)
  reduction : boolean whether this layer is a reduction cell
  reduction_prev : boolean of whether the last cell was a reduction cell
  """
  def __init__(self, genotype, C_prev_prev, C_prev, C, reduction, reduction_prev):
    super(Cell, self).__init__()

    if reduction_prev:
      self.preprocess0 = FactorizedReduce(C_prev_prev, C)
    else:
      self.preprocess0 = ReLUConvBN(C_prev_prev, C, 1, 1, 0)
    self.preprocess1 = ReLUConvBN(C_prev, C, 1, 1, 0)
    self._channels = [C,C_prev,C_prev_prev] 
    self._reduc = [reduction,reduction_prev] 
    if reduction:
      op_names, indices = zip(*genotype.reduce)
      concat = genotype.reduce_concat
    else:
      op_names, indices = zip(*genotype.normal) #get operation names and indexs from genotype
      concat = genotype.normal_concat
    self._compile(C, op_names, indices, concat, reduction)

  def _compile(self, C, op_names, indices, concat, reduction):
    assert len(op_names) == len(indices)
    self._steps = len(op_names) // 2
    self._concat = concat
    self.multiplier = len(concat)

    self._ops = nn.ModuleList()
    for name, index in zip(op_names, indices):
      stride = 2 if reduction and index < 2 else 1
      op = OPS[name](C, stride, True)
      self._ops += [op]
    self._indices = indices

  def forward(self, s0, s1, drop_prob):
    s0 = self.preprocess0(s0)
    s1 = self.preprocess1(s1)
    
    states = [s0, s1]
    for i in range(self._steps):
      h1 = states[self._indices[2*i]]
      h2 = states[self._indices[2*i+1]]
      op1 = self._ops[2*i]
      op2 = self._ops[2*i+1]
    
      h1 = op1(h1)
      h2 = op2(h2)
      if self.training and drop_prob > 0.:
        if not isinstance(op1, Identity):
          h1 = drop_path(h1, drop_prob)
        if not isinstance(op2, Identity):
          h2 = drop_path(h2, drop_prob)
      if h1.shape[2] != h2.shape[2]:
        print(op1,op2)
        print("Channels k, k-1, k-2: {} {} {}".format(*self._channels))
        print("Reduction: {} - Previous Reduction: {} ".format(*self._reduc))
      s = h1 + h2
      states += [s]
    return torch.cat([states[i] for i in self._concat], dim=1)


class NetworkTS(nn.Module):

  def __init__(self, stem_channels ,C, num_classes, layers, auxiliary, drop_prob ,genotype, binary = False):
    super(NetworkTS, self).__init__()
    self._layers = layers
    self.drop_path_prob = drop_prob
    #Stem Setup
    stem_multiplier = 3
    C_curr = stem_multiplier*C
    self.stem = nn.Sequential(
      nn.Conv1d(stem_channels, C_curr, 3, padding=1, bias=False),
      nn.BatchNorm1d(C_curr)
    )
        
    C_prev_prev, C_prev, C_curr = C_curr, C_curr, C #Initialising channels
    self.cells = nn.ModuleList()
    reduction_prev = False
    
    for i in range(layers):
      if i %2 ==0 and i != layers: # Set as reduction layer if model is 1/3 or 2/3 through
                                        # the network
        C_curr *= 2 # double channels on reduction layer
        reduction = True
      else:
        reduction = False
      #Build cell
      cell = CellAttn(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
      reduction_prev = reduction
      self.cells += [cell]
      C_prev_prev, C_prev = C_prev, cell.multiplier*C_curr # Update channels
      if i == 2*layers//3:
        C_to_auxiliary = C_prev

    self.global_pooling = nn.AdaptiveAvgPool1d(1)
    self.c_out = C_prev
    if binary == True:
      self.binary = binary
      self.classifier = nn.Linear(C_prev, 1)
    else:
      self.classifier = nn.Linear(C_prev, num_classes)
  
  def get_channels(self):
    return self.c_out
  def _forward(self, input):
    logits_aux = None
    s0 = s1 = self.stem(input)
    for i, cell in enumerate(self.cells):
      s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
    out = self.global_pooling(s1)
    return out.view(out.size(0),-1)

  def forward(self, input):
    out = self._forward(input)
    logits = self.classifier(out)
    #if self.binary:
    #  logits = self.outact(logits)
    return logits


class NetworkMain(nn.Module):

  def __init__(self, stem_channels ,C, num_classes, layers, auxiliary, drop_prob ,genotype, binary = False):
    super(NetworkMain, self).__init__()
    self._layers = layers
    self._auxiliary = auxiliary
    self.drop_path_prob = drop_prob
    #Stem Setup
    stem_multiplier = 3
    C_curr = stem_multiplier*C
    self.stem = nn.Sequential(
      nn.Conv1d(stem_channels, C_curr, 3, padding=1, bias=False),
      nn.BatchNorm1d(C_curr)
    )
        
    C_prev_prev, C_prev, C_curr = C_curr, C_curr, C #Initialising channels
    self.cells = nn.ModuleList()
    reduction_prev = False
    
    for i in range(layers):
      if i %2 ==0 and i != layers: # Set as reduction layer if model is 1/3 or 2/3 through
                                        # the network
        C_curr *= 2 # double channels on reduction layer
        reduction = True
      else:
        reduction = False
      #Build cell
      cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
      reduction_prev = reduction
      self.cells += [cell]
      C_prev_prev, C_prev = C_prev, cell.multiplier*C_curr # Update channels
      if i == 2*layers//3:
        C_to_auxiliary = C_prev

    if auxiliary:
      self.auxiliary_head = AuxiliaryHeadCIFAR(C_to_auxiliary, num_classes)
    self.global_pooling = nn.AdaptiveAvgPool1d(1)
    self.c_out = C_prev
    if binary == True:
      self.binary = binary
      #self.outact = nn.Sigmoid()
      self.classifier = nn.Linear(C_prev, 1)
    else:
      self.classifier = nn.Linear(C_prev, num_classes)
  
  def get_channels(self):
    return self.c_out
  def _forward(self, input):
    logits_aux = None
    s0 = s1 = self.stem(input)
    for i, cell in enumerate(self.cells):
      s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
      if i == 2*self._layers//3:
        if self._auxiliary and self.training:
          logits_aux = self.auxiliary_head(s1)
    out = self.global_pooling(s1)
    return out.view(out.size(0),-1)

  def forward(self, input):
    out = self._forward(input)
    logits = self.classifier(out)
    #if self.binary:
    #  logits = self.outact(logits)
    return logits





class AuxiliaryHeadCIFAR(nn.Module):

  def __init__(self, C, num_classes):
    """assuming input size 8x8"""
    super(AuxiliaryHeadCIFAR, self).__init__()
    self.features = nn.Sequential(
      nn.ReLU(inplace=True),
      nn.AvgPool2d(5, stride=3, padding=0, count_include_pad=False), # image size = 2 x 2
      nn.Conv2d(C, 128, 1, bias=False),
      nn.BatchNorm2d(128),
      nn.ReLU(inplace=True),
      nn.Conv2d(128, 768, 2, bias=False),
      nn.BatchNorm2d(768),
      nn.ReLU(inplace=True)
    )
    self.classifier = nn.Linear(768, num_classes)

  def forward(self, x):
    x = self.features(x)
    x = self.classifier(x.view(x.size(0),-1))
    return x


class AuxiliaryHeadImageNet(nn.Module):

  def __init__(self, C, num_classes):
    """assuming input size 14x14"""
    super(AuxiliaryHeadImageNet, self).__init__()
    self.features = nn.Sequential(
      nn.ReLU(inplace=True),
      nn.AvgPool2d(5, stride=2, padding=0, count_include_pad=False),
      nn.Conv2d(C, 128, 1, bias=False),
      nn.BatchNorm2d(128),
      nn.ReLU(inplace=True),
      nn.Conv2d(128, 768, 2, bias=False),
      # NOTE: This batchnorm was omitted in my earlier implementation due to a typo.
      # Commenting it out for consistency with the experiments in the paper.
      # nn.BatchNorm2d(768),
      nn.ReLU(inplace=True)
    )
    self.classifier = nn.Linear(768, num_classes)

  def forward(self, x):
    x = self.features(x)
    x = self.classifier(x.view(x.size(0),-1))
    return x


class NetworkCIFAR(nn.Module):

  def __init__(self, C, num_classes, layers, drop_prob, auxiliary, genotype):
    super(NetworkCIFAR, self).__init__()
    self.drop_path_prob = drop_prob
    self._layers = layers
    self._auxiliary = auxiliary

    stem_multiplier = 3
    C_curr = stem_multiplier*C
    self.stem = nn.Sequential(
      nn.Conv2d(3, C_curr, 3, padding=1, bias=False),
      nn.BatchNorm2d(C_curr)
    )
    
    C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
    self.cells = nn.ModuleList()
    reduction_prev = False
    for i in range(layers):
      if i in [layers//3, 2*layers//3]:
        C_curr *= 2
        reduction = True
      else:
        reduction = False
      cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
      reduction_prev = reduction
      self.cells += [cell]
      C_prev_prev, C_prev = C_prev, cell.multiplier*C_curr
      if i == 2*layers//3:
        C_to_auxiliary = C_prev

    if auxiliary:
      self.auxiliary_head = AuxiliaryHeadCIFAR(C_to_auxiliary, num_classes)
    self.global_pooling = nn.AdaptiveAvgPool2d(1)
    self.classifier = nn.Linear(C_prev, num_classes)


  def forward(self, input):
    logits_aux = None
    s0 = s1 = self.stem(input)
    for i, cell in enumerate(self.cells):
      s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
      if i == 2*self._layers//3:
        if self._auxiliary and self.training:
          logits_aux = self.auxiliary_head(s1)
    out = self.global_pooling(s1)
    logits = self.classifier(out.view(out.size(0),-1))
    if self._auxiliary:
      return logits, logits_aux
    else:
      return logits

class NetworkImageNet(nn.Module):

  def __init__(self, C, num_classes, layers, auxiliary, genotype):
    super(NetworkImageNet, self).__init__()
    self._layers = layers
    self._auxiliary = auxiliary

    self.stem0 = nn.Sequential(
      nn.Conv2d(3, C // 2, kernel_size=3, stride=2, padding=1, bias=False),
      nn.BatchNorm2d(C // 2),
      nn.ReLU(inplace=True),
      nn.Conv2d(C // 2, C, 3, stride=2, padding=1, bias=False),
      nn.BatchNorm2d(C),
    )

    self.stem1 = nn.Sequential(
      nn.ReLU(inplace=True),
      nn.Conv2d(C, C, 3, stride=2, padding=1, bias=False),
      nn.BatchNorm2d(C),
    )

    C_prev_prev, C_prev, C_curr = C, C, C

    self.cells = nn.ModuleList()
    reduction_prev = True
    for i in range(layers):
      if i in [layers // 3, 2 * layers // 3]:
        C_curr *= 2
        reduction = True
      else:
        reduction = False
      cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
      reduction_prev = reduction
      self.cells += [cell]
      C_prev_prev, C_prev = C_prev, cell.multiplier * C_curr
      if i == 2 * layers // 3:
        C_to_auxiliary = C_prev

    if auxiliary:
      self.auxiliary_head = AuxiliaryHeadImageNet(C_to_auxiliary, num_classes)
    self.global_pooling = nn.AvgPool2d(7)
    self.classifier = nn.Linear(C_prev, num_classes)

  def forward(self, input):
    logits_aux = None
    s0 = self.stem0(input)
    s1 = self.stem1(s0)
    for i, cell in enumerate(self.cells):
      s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
      if i == 2 * self._layers // 3:
        if self._auxiliary and self.training:
          logits_aux = self.auxiliary_head(s1)
    out = self.global_pooling(s1)
    logits = self.classifier(out.view(out.size(0), -1))
    return logits, logits_aux

